/* ******************************************************************* *\
*                               datetime.d                              *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*       Data types to implement object-oriented date manipulation       *
>***********************************************************************<
* Copyright (C) 2005-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/*!	The smjg.libs.util.datetime module provides an object-oriented means of manipulating dates and
 *	times.
 *
 *	Todo:
 *		Support interrogation of current date/time and time zone on Linux and other platforms.
 *		Add more features (see readme).
 */
module smjg.libs.util.datetime;

private import
	smjg.libs.util.datetimeformat;
private import std.math:   floor;
private import std.string: format;


///	Constants representing the days of the week.
enum DAY_OF_WEEK : byte {
	UNSPEC = byte.min, SUN = 0, MON, TUE, WED, THU, FRI, SAT }
///	Constants representing the months of the year.
enum MONTH : byte {
	UNSPEC = byte.min,
	JAN = 1, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
}

///	Short (three-letter) names of the days of the week.
immutable string[MONTH.max + 1] SHORT_DAY_NAME = [
	DAY_OF_WEEK.SUN: "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
];

///	Full names of the days of the week.
immutable string[MONTH.max + 1] LONG_DAY_NAME = [
	DAY_OF_WEEK.SUN: "Sunday", "Monday", "Tuesday", "Wednesday",
	  "Thursday", "Friday", "Saturday"
];

///	Short (three-letter) names of the months of the year.
immutable string[MONTH.max + 1] SHORT_MONTH_NAME = [
	['\xFF', '\xFF', '\xFF'],
	MONTH.JAN: "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
];

///	Full names of the months of the year.
immutable string[MONTH.max + 1] LONG_MONTH_NAME = [
	null, MONTH.JAN: "January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December"
];

///	The length of each month in a common year.
const int MONTH_LEN[MONTH.max + 1] = [
	MONTH.JAN: 31, MONTH.FEB: 28, MONTH.MAR: 31, MONTH.APR: 30,
	MONTH.MAY: 31, MONTH.JUN: 30, MONTH.JUL: 31, MONTH.AUG: 31,
	MONTH.SEP: 30, MONTH.OCT: 31, MONTH.NOV: 30, MONTH.DEC: 31
];

/*	The number of days from 2 January to the first of each month in a common
 *	year.  This is used to convert dates to date values.
 */
private const int MONTH_OFFSET[MONTH.max + 1] = [
	MONTH.JAN:  -1, MONTH.FEB:  30, MONTH.MAR:  58, MONTH.APR:  89,
	MONTH.MAY: 119, MONTH.JUN: 150, MONTH.JUL: 180, MONTH.AUG: 211,
	MONTH.SEP: 242, MONTH.OCT: 272, MONTH.NOV: 303, MONTH.DEC: 333
];

private const int UNIX_EPOCH = 719527;

///	Determines whether the given year is a leap year.
bool isLeapYear(int year) {
	return year % 4 == 0 && year % 100 != 0 || year % 400 == 0;
}


/**	The local time zone offset from UTC in minutes.
 *
 *	$(P A positive value indicates a zone ahead of UTC; a negative value behind.
 *	Under Windows, this is automatically initialised to the system time zone.)
 *
 *	Todo: implement initialisation on platforms other than Windows.
 */
int timeZoneOffset;

/**	Whether to allow alignment fields to overflow.
 *
 *	$(P If this is false, a DateTimeFormatException will be thrown if the formatted
 *	content of a field exceeds the field's specified width.)
 */
bool allowFieldOverflow = false;


version (Windows) {
	import std.c.windows.windows;

	extern (Windows) {
		int GetDateFormatA(uint Locale, DWORD dwFlags, SYSTEMTIME* lpDate,
		  LPCSTR lpFormat, LPSTR lpDateStr, int cchDate);
		int GetTimeFormatA(uint Locale, DWORD dwFlags, SYSTEMTIME* lpTime,
		  LPCSTR lpFormat, LPSTR lpTimeStr, int cchTime);
	}

	static this() {
		TIME_ZONE_INFORMATION tzi;
		uint dst = GetTimeZoneInformation(&tzi);

		timeZoneOffset = -tzi.Bias;
		switch (dst) {
			case TIME_ZONE_ID_UNKNOWN:
				break;
			case TIME_ZONE_ID_STANDARD:
				timeZoneOffset -= tzi.StandardBias;
				break;
			case TIME_ZONE_ID_DAYLIGHT:
				timeZoneOffset -= tzi.DaylightBias;
				break;
			default:
				assert(false, "GetTimeZoneInformation returned unrecognised DST value");
		}

		debug (datetime) with(tzi) {
			writefln("Base timezone offset:       %d", Bias);
			writefln("Standard time adjustment:   %d", StandardBias);
			writefln("Daylight saving adjustment: %d", DaylightBias);
		}
	}
}


/**	Determines the length of a month.
 *
 *	Params:
 *		month = month whose length is to be found.
 *		year  = year in which the month occurs.
 *	Returns:
 *		the length of the given month in days.
 */
int monthLength(int month, int year)
in {
	assert (month >= 1);
}
body {
	if (month == 2 && isLeapYear(year)) {
		return 29;
	} else {
		return MONTH_LEN[month];
	}
}


/*	Determines the offset of each month, used to calculate date values.
 *
 *	Params:
 *		month = month whose offset is to be found.
 *		year  = year in which the month occurs.
 *	Returns:
 *		the offset of the first day of the month from 2 January in the same year.
 */
private int monthOffset(int month, int year)
in {
	assert (month >= 1);
}
body {
	return MONTH_OFFSET[month] - (month < 3 && isLeapYear(year));
}


private template Interval(T) {
	int value;
	alias value this;
	this(int v) { value = v; }

	T opAdd(T ti) const {
		return T(value + ti.value);
	}
	T opSub(T ti) const {
		return T(value - ti.value);
	}
	T opNeg() const {
		return T(-value);
	}
	
	T opMul(int v) const {
		return T(value * v);
	}
	T opDiv(int v) const {
		return T(value / v);
	}
	int opDiv(T ti) const {
		return value / ti.value;
	}
	T opMod(T ti) const {
		return T(value % ti.value);
	}
}


///	The DateInterval structure represents an interval between date values, specified in days.
struct DateInterval {
	mixin Interval!(DateInterval) Ops;
	alias Ops.opAdd opAdd;
	alias Ops.opSub opSub;
	
	DateTimeInterval opAdd(TimeInterval ti) const {
		return DateTimeInterval(this, ti);
	}
	
	DateTimeInterval opSub(TimeInterval ti) const {
		return DateTimeInterval(this, -ti);
	}

	enum : DateInterval {
		YEAR     = DateInterval(   365),  /// a common (non-leap) year
		YEAR_4   = DateInterval(  1461),  /** a period of four years, assumed not to include the
		                                      night of 28 February-1 March in a centennial common
		                                      year */
		YEAR_100 = DateInterval( 36524),  /** a century, assumed not to include 29 February in a
		                                      quatercentennial leap year */
		YEAR_400 = DateInterval(146097)   /// a period of 400 years.
	}
}

/**	The TimeInterval structure represents an interval between time values, specified in
 *	milliseconds.
 */
struct TimeInterval {
	mixin Interval!(TimeInterval) Ops;
	alias Ops.opAdd opAdd;
	alias Ops.opSub opSub;

	DateTimeInterval opAdd(DateInterval di) const {
		return DateTimeInterval(di, this);
	}
	DateTimeInterval opSub(DateInterval di) const {
		return DateTimeInterval(-di, this);
	}

	enum : TimeInterval {
		MS     = TimeInterval       (1), /// Constants representing units of time.
		SECOND = TimeInterval    (1000), /// ditto
		MINUTE = TimeInterval   (60000), /// ditto
		HOUR   = TimeInterval (3600000), /// ditto
		DAY    = TimeInterval(86400000)  /// ditto
	}
}


///	The DateTimeInterval structure represents an interval in both date and time.
struct DateTimeInterval {
	///	The date component of the interval in days.
	DateInterval datePart;
	/**	The time component of the interval in milliseconds.
	 *
	 *	$(P In a normalised interval, this value will be in the interval
	 *	[0..TimeInterval.DAY$(RPAREN).)
	 */
	TimeInterval timePart;

	/**	Constructs a DateTimeInterval from the given date and time intervals.
	 *
	 *	Params:
	 *		days   = date interval in days.
	 *		millis = time interval in milliseconds.
	 */
	this(int days, int millis) {
		datePart = DateInterval(days);
		timePart = TimeInterval(millis);
	}

	///	Constructs a DateTimeInterval equal to a given date interval.
	this(DateInterval days) {
		datePart = days;
		timePart = 0;
	}

	///	Constructs a DateTimeInterval equal to a given time interval.
	this(TimeInterval millis) {
		datePart = 0;
		timePart = millis;
	}

	///	Constructs a DateTimeInterval equal to a given value in milliseconds.
	this(long millis) {
		datePart = DateInterval(cast(int) (millis / TimeInterval.DAY.value));
		timePart = TimeInterval(cast(int) (millis % TimeInterval.DAY.value));
		if (timePart < 0) {
			datePart -= 1;
			timePart += TimeInterval.DAY;
		}
	}

	/**	Normalises this DateTimeInterval in place.
	 *
	 *	$(P The $(D timePart) will be normalised to the range $(D [0..TimeInterval.DAY&#41;), and
	 *	$(D datePart) will be adjusted accordingly.)
	 *
	 *	$(P All methods of DateTimeInterval (except for adding or subtracting a DateInterval from
	 *	an already unnormalised DateTimeInterval) return a normalised interval where applicable.
	 *	This method is provided so that you can normalise the interval after modifying timePart
	 *	directly.)
	 *
	 *	Returns: this interval.
	 */
	@property DateTimeInterval normalise() {
		if (timePart < 0) {
			datePart += timePart / TimeInterval.DAY - 1;
			timePart = timePart % TimeInterval.DAY + TimeInterval.DAY;
		} else if (timePart >= TimeInterval.DAY) {
			datePart += timePart / TimeInterval.DAY;
			timePart %= TimeInterval.DAY;
		}
		return this;
	}

	///	The value of this DateTimeInterval in milliseconds.
	@property long milliseconds() const {
		return cast(long) datePart * TimeInterval.DAY.value + timePart.value;
	}
	///
	@property int hour() const { return timePart / TimeInterval.HOUR; }
	///
	@property int minute() const { return (timePart % TimeInterval.HOUR) / TimeInterval.MINUTE; }
	///
	@property int second() const { return (timePart % TimeInterval.MINUTE) / TimeInterval.SECOND; }
	///
	@property int millisecond() const { return timePart % TimeInterval.SECOND; }

	///	Returns a string representation in the form e.g. "10 days 12:34:56.789".
	@property string toString() const {
		return std.string.format("%d days %02d:%02d:%02d.%03d",
		  cast(int) datePart, hour, minute, second, millisecond);
	}

	///	Since: 0.10
	bool opEquals(DateTimeInterval di) const {
		return this is di || milliseconds == di.milliseconds;
	}

	///	Since: 0.10
	int opCmp(DateTimeInterval di) const {
		long ms1 = milliseconds;
		long ms2 = di.milliseconds;
		if (ms1 < ms2) return -1;
		if (ms1 > ms2) return 1;
		return 0;
	}

	DateTimeInterval opAdd(DateTimeInterval di) const {
		return DateTimeInterval(datePart + di.datePart,
		  timePart + di.timePart).normalise;
	}

	DateTimeInterval opAdd(DateInterval di) const {
		DateTimeInterval result = this;
		result.datePart += di;
		return result;
	}

	DateTimeInterval opAdd(TimeInterval ti) const {
		return opAdd(DateTimeInterval(ti));
	}

	DateTimeInterval opSub(DateTimeInterval di) const {
		return DateTimeInterval(datePart - di.datePart,
		  timePart - di.timePart);
	}

	DateTimeInterval opSub(DateInterval di) const {
		DateTimeInterval result = this;
		result.datePart -= di;
		return result;
	}

	DateTimeInterval opSub(TimeInterval ti) const {
		return opSub(DateTimeInterval(ti));
	}

	DateTimeInterval opMul(int i) const {
		long timeMul = cast(long) timePart * i;

		DateTimeInterval result;
		result.timePart = TimeInterval(timeMul % TimeInterval.DAY.value);
		result.datePart = DateInterval(cast(int)
		  (datePart.value * i + timeMul / TimeInterval.DAY.value));
		if (result.timePart < 0) {
			result.timePart += TimeInterval.DAY;
			result.datePart -= 1;
		}
		return result;
	}

	DateTimeInterval opMul(real r) const {
		long newMs = cast(long) (milliseconds * r);
		return DateTimeInterval(newMs);
	}

	DateTimeInterval opDiv(int i) const {
		return DateTimeInterval(cast(long) (milliseconds / i));
	}

	DateTimeInterval opDiv(real r) const {
		long newMs = cast(long) (milliseconds / r);
		return DateTimeInterval(newMs);
	}

	real opDiv(DateTimeInterval dti) const {
		return cast(real) milliseconds / dti.milliseconds;
	}

	DateTimeInterval opAddAssign(DateInterval di) {
		datePart += di;
		return this;
	}

	DateTimeInterval opAddAssign(TimeInterval ti) {
		return opAddAssign(DateTimeInterval(ti));
	}

	DateTimeInterval opSubAssign(DateInterval di) {
		datePart -= di;
		return this;
	}

	DateTimeInterval opSubAssign(TimeInterval ti) {
		return opSubAssign(DateTimeInterval(ti));
	}

	DateTimeInterval opAddAssign(DateTimeInterval di) { return this = this + di; }
	DateTimeInterval opSubAssign(DateTimeInterval di) { return this = this - di; }
	DateTimeInterval opMulAssign(int i) { return this = this * i; }
	DateTimeInterval opMulAssign(real r) { return this = this * r; }
	DateTimeInterval opDivAssign(int i) { return this = this / i; }
	DateTimeInterval opDivAssign(real r) { return this = this / r; }

	unittest {
		import std.stdio;

		writefln("Unittest commenced at %s", DateTime.now.toString);

		DateTimeInterval dti
		  = DateTimeInterval(10, 7*TimeInterval.HOUR + 13*TimeInterval.MINUTE);
		assert (dti.timePart == 25980000);
		assert ((dti + DateTimeInterval(2, 18*TimeInterval.HOUR)).toString
		  == "13 days 01:13:00.000");

		assert ((dti + cast(DateInterval) 12).toString
		  == "22 days 07:13:00.000");
		assert ((dti + cast(TimeInterval) 123456).toString
		  == "10 days 07:15:03.456");

		assert ((dti * 2).toString == "20 days 14:26:00.000");
		assert ((dti * 3).toString == "30 days 21:39:00.000");
		assert ((dti * 4).toString == "41 days 04:52:00.000");
		assert ((dti * -2).toString == "-21 days 09:34:00.000");

		dti += cast(DateInterval) 2;
		assert (dti.toString == "12 days 07:13:00.000");
		dti *= 3;
		assert (dti.toString == "36 days 21:39:00.000");
		dti *= 2;
		assert (dti.toString == "73 days 19:18:00.000");
		dti /= 10;
		assert (dti.toString == "7 days 09:07:48.000");
		assert ((dti * 2.4).toString == "17 days 17:06:43.200");

		DateTime dt1 = DateTime(1979, 9, 24, 3, 5, 20, 123);
		assert ((dt1 + cast(DateInterval) 365).toString
		  == "Tue 23 Sep 1980 03:05:20.123");
		assert ((dt1 + cast(TimeInterval) 4000000).toString
		  == "Mon 24 Sep 1979 04:12:00.123");

		DateTime dt2 = dt1 + DateTimeInterval(365, 4000000);
		assert (dt2.toString == "Tue 23 Sep 1980 04:12:00.123");

		dti = (dt2 + cast(DateInterval) 1) - dt1;
		assert (dti.toString == "366 days 01:06:40.000");

		dt2 = dt1 - dti;
		assert (dt2.toString == "Sat 23 Sep 1978 01:58:40.123");
	}
}


///	The DateValue structure represents a calendar date.
struct DateValue {
	/**	The internal numerical representation of the date, in days from 1 January 1BC (as it would
	 *	be defined in the Gregorian calendar).
	 *
	 *	$(P A value of $(D int.min) denotes an invalid or null date.)
	 */
	int dayNo = int.min;

	///	Deprecated: now defined in DateInterval
	deprecated alias DateInterval.YEAR     YEAR;
	///	ditto
	deprecated alias DateInterval.YEAR_4   YEAR_4;
	///	ditto
	deprecated alias DateInterval.YEAR_100 YEAR_100;
	///	ditto
	deprecated alias DateInterval.YEAR_400 YEAR_400;
	

	///	Constructs a DateValue with the given dayNo value.
	this(int day) {
		dayNo = day;
	}

	/**	Constructs a DateValue representing the given date.
	 *
	 *	Params:
	 *		year  = the year.
	 *		month = the month, from 1 (January) to 12 (December).  A MONTH value can be used.
	 *		day   = the day of the month.
	 *	Throws:
	 *		DateTimeException if month or day is out of range.
	 *
	 *	Bugs: will not work with dates before 1BC.
	 */
	this(int year, int month, int day) {
		if (month < 1 || month > 12
			  || day < 1 || day > monthLength(month, year)) {
			throw new DateTimeException(
			  std.string.format("Invalid date %04d-%02d-%02d",
			    year, month, day));
		}

		dayNo = (year * DateInterval.YEAR.value) + (year / 4) - (year / 100) + (year / 400)
		  + monthOffset(month, year) + day;
		static if ((-42) / 10 == -4) {
			/*	compensate for behaviour of division on -ve numbers
			 *	given here as desired/actual
			year         year/4   -(year/100)  year/400      total (exc year term)
			-401       -101/-100      5/4        -2/-1      -98/-97
			-400       -100/-100      4/4        -1/-1      -97/-97
			-399       -100/ -99      4/3        -1/ 0      -97/-96
			-101        -26/ -25      2/1        -1/ 0      -25/-24
			-100        -25/ -25      1/1        -1/ 0      -25/-24
			 -99        -25/ -24      1/0        -1/ 0      -25/-24
			  -5         -2/  -1      1/0        -1/ 0       -2/ -1
			  -4         -1/  -1      1/0        -1/ 0       -1/ -1
			  -3         -1/   0      1/0        -1/ 0       -1/  0
			  -2         -1/   0      1/0        -1/ 0       -1/  0
			  -1         -1/   0      1/0        -1/ 0       -1/  0
			   0          0/   0      0/0         0/ 0        0/  0
			   1          0/   0      0/0         0/ 0        0/  0
			   2          0/   0      0/0         0/ 0        0/  0
			   3          0/   0      0/0         0/ 0        0/  0
			   4          1/   1      0/0         0/ 0        1/  1
			 */
			if (year < 0 && !isLeapYear(year)) dayNo--;
		}
	}
	//@}

	version (Windows) {
		/**	Retrieves the current system date.
		 *
		 *	Bugs: always gets the date in the system time zone, even if timeZoneOffset has been
		 *	      overridden.
		 *	Todo: Implement on platforms other than Windows.
		 */
		static @property DateValue now() {
			SYSTEMTIME st;
			GetLocalTime(&st);
			return DateValue(st.wYear, st.wMonth, st.wDay);
		}
	}

	///	Calculates the nearest w to this date.
	DateValue nearest(DAY_OF_WEEK w) const {
		int offset = w - dayOfWeek;
		if (offset >= 4) {
			return opAdd(offset) - 7;
		} else if (offset <= -4) {
			return opAdd(offset) + 7;
		} else {
			return opAdd(offset);
		}
	}

	/** Formats this date according to a _format string.
	 *
	 *	Params:
	 *		formatString = the _format string, as defined in smjg.libs.util.datetimeformat.
	 *	Returns:
	 *		the formatted date string.
	 *	Throws:
	 *		DateTimeFormatException if the formatting fails, e.g. because of an error in the
	 *	    _format string.
	 */
	string format(string formatString) const {
		DateTimeComponents dtc = DateTimeComponents(this);
		return formatDateTime(&dtc, dayOfWeek, formatString);
	}
	///
	@property int year() const {
		int dummy;

		return extractYear(dummy);
	}
	///
	@property MONTH month() const {
		int d;
		bool leap = isLeapYear(extractYear(d));

		return extractMonthDay(d, leap);
	}
	///
	@property int day() const {
		int m, d;
		bool leap = isLeapYear(extractYear(d));

		extractMonthDay(d, leap);
		return d;
	}
	///
	@property DAY_OF_WEEK dayOfWeek() const {
		return cast(DAY_OF_WEEK) ((dayNo % 7 + 7) % 7);
	}

	///	Returns a string representation in the form e.g. "Sat 8 May 2004".
	@property string toString() const {
		int d;
		int y = extractYear(d);
		int m = extractMonthDay(d, isLeapYear(y));

		if (y <= 0) {
			return std.string.format("%s %d %s %d BC",
			  SHORT_DAY_NAME[dayOfWeek], d, SHORT_MONTH_NAME[m], 1 - y);
		} else {
			return std.string.format("%s %d %s %d",
			  SHORT_DAY_NAME[dayOfWeek], d, SHORT_MONTH_NAME[m], y);
		}
	}

	version (Windows) {
		/**	Formats this date according to the OS-defined format.
		 *
		 *	Params:
		 *		flags = a value to pass to the OS date-formatting API determining how to format
		 *		the date.
		 *	Returns:
		 *		a string representation of this date in the OS-defined format.
		 *
		 *	Todo: implement on platforms other than Windows.
		 */
		char[] toFormatString(uint flags = 1) const {
			char[32] buf;
			SYSTEMTIME st;

			st.wYear = cast(ushort) year;
			if (st.wYear != year) {
				throw new DateTimeException("Date/time system formatting failed: date out of range");
			}
			st.wMonth = month;
			st.wDayOfWeek = dayOfWeek;
			st.wDay = cast(ushort) day;

			int len = GetDateFormatA(0, flags, &st, null, buf.ptr, buf.length);
			return buf[0..len].dup;
		}

		/**	Formats this date according to the OS-defined short format.
		 *
		 *	Returns: a string representation of the date in the OS-defined short format.
		 *
		 *	Todo: Implement on platforms other than Windows.
		 */
		@property char[] toShortFormatString() const {
			return toFormatString(1);
		}

		/**	Formats this date according to the OS-defined long format.
		 *
		 *	Returns: a string representation of the date in the OS-defined long format.
		 *
		 *	Todo: Implement on platforms other than Windows.
		 */
		@property char[] toLongFormatString() const {
			return toFormatString(2);
		}
	}

	///	Adds the given interval in days to this date.
	DateValue opAdd(int i) const {
		return DateValue(dayNo + i);
	}

	///	Subtracts the given interval in days from this date.
	DateValue opSub(int i) const {
		return DateValue(dayNo - i);
	}

	///	Calculates the interval in days between this date and d.
	DateInterval opSub(DateValue d) const {
		return cast(DateInterval) (dayNo - d.dayNo);
	}

	///	Adds the given interval in days to this date.
	DateValue opAddAssign(int i) {
		dayNo += i;
		return this;
	}

	///	Subtracts the given interval in days from this date.
	DateValue opSubAssign(int i) {
		dayNo -= i;
		return this;
	}

	alias opSub opCmp;

	/*	Extracts the year from this date object.
	 *
	 *	Params:
	 *		doy = variable to receive the number of the day within the year for passing to
	 *		      extractMonthDay.
	 */
	private int extractYear(out int doy) const {
		int result = 0;
		enum : int {
			YEAR     = DateInterval.YEAR.value,
			YEAR_4   = DateInterval.YEAR_4.value,
			YEAR_100 = DateInterval.YEAR_100.value,
			YEAR_400 = DateInterval.YEAR_400.value
		}
		doy = dayNo;

		if (doy < 0) {	// date prior to 1BC
			result = 400 * (doy / YEAR_400 - 1);
			doy = doy % YEAR_400 + YEAR_400;
		} else {
			result += 400 * (doy / YEAR_400);
			doy %= YEAR_400;
		}
		debug (datetime) writefln("%d %d", result, doy);

		if (doy == YEAR_100 * 4) {
			// date is 1 January of a 400th year
			debug (datetime) writefln("1 January quatercentennial detected");
			doy = -1;
			return result + 400;
		}

		result += 100 * (doy / YEAR_100);
		doy %= YEAR_100;
		debug (datetime) writefln("%d %d", result, doy);

		result += 4 * (doy / YEAR_4);
		doy %= YEAR_4;
		debug (datetime) writefln("%d %d", result, doy);

		if (doy == YEAR * 4) {
			// date is 1 January of a leap year
			debug (datetime) writefln("1 January leap detected");
			result += 4;
			doy = -1;
		} else {
			result += doy / YEAR;
			doy %= YEAR;
		}
		debug (datetime) writefln("%d %d", result, doy);

		return result;
	}

	unittest {
		import std.stdio;

		writefln("Unittest commenced on %s", DateValue.now.toString);
		assert (DateValue(1979, 9, 24).toString == "Mon 24 Sep 1979");

		// year rollover - common to common
		DateValue d = DateValue(2005, 12, 31);
		assert (d.toString == "Sat 31 Dec 2005");
		assert ((d+1).toString == "Sun 1 Jan 2006");
		assert ((d-365).toString == "Fri 31 Dec 2004");
		assert ((d-364).toString == "Sat 1 Jan 2005");
		assert ((d-363).toString == "Sun 2 Jan 2005");

		// year rollover - leap to common
		d = DateValue(2004, 12, 31);
		assert (d.toString == "Fri 31 Dec 2004");
		assert ((d+1).toString == "Sat 1 Jan 2005");
		assert ((d+2).toString == "Sun 2 Jan 2005");

		// year rollover - common to leap
		d = DateValue(2003, 12, 31);
		assert (d.toString == "Wed 31 Dec 2003");
		assert ((d+1).toString == "Thu 1 Jan 2004");
		assert ((d+2).toString == "Fri 2 Jan 2004");

		// year rollover - common to centennial common
		d = DateValue(1899, 12, 31);
		assert (d.toString == "Sun 31 Dec 1899");
		assert ((d+1).toString == "Mon 1 Jan 1900");
		assert ((d+2).toString == "Tue 2 Jan 1900");

		// year rollover - common to quatercentennial leap
		d = DateValue(1999, 12, 31);
		assert (d.toString == "Fri 31 Dec 1999");
		assert ((d+1).toString == "Sat 1 Jan 2000");
		assert ((d+2).toString == "Sun 2 Jan 2000");

		// month rollover
		d = DateValue(1979, 9, 24) + 6;
		assert (d.toString == "Sun 30 Sep 1979");
		d += 1;
		assert (d.toString == "Mon 1 Oct 1979");
		// add days into leap year
		d += 365;
		assert (d.toString == "Tue 30 Sep 1980");

		// more year rollover
		d = DateValue(1979, 12, 25) + 6;
		assert (d.toString == "Mon 31 Dec 1979");
		d += 1;
		assert (d.toString == "Tue 1 Jan 1980");
		assert ((d+1).toString == "Wed 2 Jan 1980");
		assert ((d+2).toString == "Thu 3 Jan 1980");

		// February-March rollover in leap year
		d += 30;
		assert (d.toString == "Thu 31 Jan 1980");
		d += 1;
		assert (d.toString == "Fri 1 Feb 1980");
		d += 28;
		assert (d.toString == "Fri 29 Feb 1980");
		assert ((d+1).toString == "Sat 1 Mar 1980");
		assert ((d+2).toString == "Sun 2 Mar 1980");

		// February-March rollover in common year
		d = DateValue(1979, 1, 1);
		assert (d.toString == "Mon 1 Jan 1979");
		d += 30;
		assert (d.toString == "Wed 31 Jan 1979");
		d += 1;
		assert (d.toString == "Thu 1 Feb 1979");
		d += 27;
		assert (d.toString == "Wed 28 Feb 1979");
		assert ((d+1).toString == "Thu 1 Mar 1979");
		assert ((d+2).toString == "Fri 2 Mar 1979");

		// nearest day of week calculations
		assert (DateValue(1979, 9, 24).nearest(DAY_OF_WEEK.WED)
		  == DateValue(1979, 9, 26));
		assert (DateValue(1982, 8, 15).nearest(DAY_OF_WEEK.FRI)
		  == DateValue(1982, 8, 13));

		/*	test BC dates as well
		 *	(this is long before the Gregorian Calendar was conceived, but for
		 *	consistency's sake it should work
		 */
		d = DateValue(1, 9, 24);
		assert (d.toString == "Mon 24 Sep 1");
		assert ((d-365).toString == "Sun 24 Sep 1 BC");
		assert ((d-730).toString == "Sat 25 Sep 2 BC");
		d = DateValue(-3, 9, 24);
		assert ((d+730).toString == "Fri 24 Sep 2 BC");
		assert (d.toString == "Wed 24 Sep 4 BC");
		assert ((d-365).toString == "Tue 24 Sep 5 BC");
		assert ((d-730).toString == "Mon 25 Sep 6 BC");
	}
}

/*!	\internal
 *
 *	Extracts the month and day from the given day of year number.
 *
 *	\param day       number of day within year, as supplied by
 *	                 DateValue.extractYear.  This value is replaced with the
 *	                 extracted number of the day within the month.
 *	\param leapYear  whether it is a leap year.
 *	\return          the extracted month.
 */
private MONTH extractMonthDay(ref int day, bool leapYear) {
	MONTH month;

	debug (datetime) writefln("day == %d; leapYear == %d", day, leapYear);

	if (day <= MONTH_OFFSET[MONTH.MAR]) {
		// January or February - leap year dependent
		if (leapYear) day++;
		month = day <= MONTH_OFFSET[MONTH.FEB] ? MONTH.JAN : MONTH.FEB;
		day -= MONTH_OFFSET[month];
		return month;
	}

	for (month = MONTH.MAR; month <= MONTH.DEC && day > MONTH_OFFSET[month];
	  month++) {
		debug (datetime) {
			writefln("Trying month %d, offset %d", cast(byte) month,
			  MONTH_OFFSET[month]);
			day = day;
		}
	}
	debug (datetime) writefln(cast(byte) month);
	//debug (datetime) writefln(cast(byte) --month);
	--month;
	day -= MONTH_OFFSET[month];
	debug (datetime) writefln(cast(byte) month);
	assert (month >= MONTH.MAR);

	return month;
}


///	The TimeValue structure represents a time within a day.
struct TimeValue {
	/**	The internal numerical representation of the time.
	 *
	 *	$(P This value is in milliseconds from 00:00 UTC.  It must be in the range 0..86401000 or
	 *	be int.min, representing an invalid or null time.)
	 */
	int utcMillis = int.min;

	invariant() {
		assert ((utcMillis >= 0 && utcMillis <= max)
			  || utcMillis == int.min);
	}
	
	enum : int {
		min = 0,
		max = (TimeInterval.DAY + TimeInterval.SECOND).value - 1
	}

	///	Deprecated: now defined in TimeInterval
	deprecated alias TimeInterval.SECOND SECOND;
	/// ditto
	deprecated alias TimeInterval.MINUTE MINUTE;
	/// ditto
	deprecated alias TimeInterval.HOUR   HOUR;
	/// ditto
	deprecated alias TimeInterval.DAY    DAY;
	
	///	Constructs a TimeValue with the given UTC millisecond value.
	this(int ms) {
		utcMillis = ms;
	}

	///	Constructs a TimeValue representing the given local time.
	this(int hour, int minute, int second = 0, int ms = 0) {
		if (hour < 0 || hour >= 24
			  || minute < 0 || minute >= 60
			  || second < 0 || second > 60
			  || ms < 0 || ms >= 1000) {
			throw new DateTimeException(
				std.string.format("Invalid time %02d:%02d:%02d.%03d",
				  hour, minute, second, ms));
		}

		utcMillis = normalise((((hour * 60
		    + minute - timeZoneOffset) * 60 + second) * 1000) + ms);
	}

	/// ditto
	this(int hour, int minute, float second) {
		if (hour < 0 || hour >= 24
			  || minute < 0 || minute >= 60
			  || second < 0 || second >= 61) {
			throw new DateTimeException(
				std.string.format("Invalid time %02d:%02d:%06.3f",
				  hour, minute, second));
		}

		utcMillis = normalise(cast(int) ((((hour * 60
		    + minute - timeZoneOffset) * 60 + second) * 1000) + 0.5));
	}

	version (Windows) {
		/**	Retrieves the current system time.
		 *
		 *	Todo: implement on platforms other than Windows.
		 */
		static @property TimeValue now() {
			SYSTEMTIME st;
			GetSystemTime(&st);

			return TimeValue((((st.wHour * 60
		      + st.wMinute) * 60 + st.wSecond) * 1000) + st.wMilliseconds);
		}
	}

	/** Formats this time according to a _format string.
	 *	
	 *	Params:
	 *		formatString = the _format string, as defined in smjg.libs.util.datetimeformat.
	 *	Returns:
	 *		the formatted date string.
	 *	Throws:
	 *		DateTimeFormatException if the formatting fails, e.g. because of an error in the
	 *		_format string.
	 */
	string format(string formatString) const {
		DateTimeComponents dtc = DateTimeComponents(this);
		return formatDateTime(&dtc, DAY_OF_WEEK.init, formatString);
	}

	///	The time in milliseconds from midnight in local time.
	@property int localMillis() const
	out (result) {
		assert (result >= 0 && result <= max);
	}
	body {
		return normalise(utcMillis + timeZoneOffset * TimeInterval.MINUTE.value);
	}

	///	ditto
	@property int localMillis(int ms)
	in {
		assert (ms >= 0 && ms <= TimeInterval.DAY.value);
	}
	out {
		assert (localMillis == ms);
	}
	body {
		utcMillis = normalise(utcMillis - timeZoneOffset * TimeInterval.MINUTE.value);
		return ms;
	}

	///
	@property int hour() const { return localMillis / TimeInterval.HOUR.value; }
	///
	@property int minute() const { return (localMillis % TimeInterval.HOUR.value) / TimeInterval.MINUTE.value; }
	///
	@property int second() const { return (localMillis % TimeInterval.MINUTE.value) / TimeInterval.SECOND.value; }
	///
	@property int millisecond() const { return localMillis % TimeInterval.SECOND.value; }

	/**	Returns the offset of the local time calendar date from the UTC date.
	 *
	 *	Returns:
	 *		-1 if the local date is the day before UTC date,
	 *	       e.g. 01:00 UTC when the time zone is UTC-02:00.
	 *		0 if the local date equals the UTC date.
	 *		1 if the local date is the day after UTC date,
	 *			e.g. 23:00 UTC when the time zone is UTC+02:00.
	 */
	@property int localDayOffset() const {
		if (utcMillis == int.min) {
			throw new DateTimeException("Uninitialised time object");
		}
		int local = utcMillis + timeZoneOffset * TimeInterval.MINUTE.value;
		return (local - normalise(local)) / TimeInterval.DAY.value;
	}

	///	Returns a string representation in the form e.g. "22:12:31.278".
	@property string toString() const {
		return std.string.format("%02d:%02d:%02d.%03d",
		  hour, minute, second, millisecond);
	}

	version (Windows) {
		/**	Formats this time according to the OS-defined format.
		 *
		 *	Params:
		 *		flags = a value to pass to the OS time-formatting API determining how to format
		 *		        the time.
		 *	Returns:
		 *		a string representation of this time in the OS-defined format.
		 *
		 *	Todo: implement on platforms other than Windows.
		 */
		char[] toFormatString(uint flags) const {
			char[32] buf;
			SYSTEMTIME st;

			st.wHour = cast(ushort) hour;
			st.wMonth = cast(ushort) minute;
			st.wSecond = cast(ushort) second;
			st.wMilliseconds = cast(ushort) millisecond;

			int len = GetTimeFormatA(0, flags, &st, null, buf.ptr, buf.length);
			return buf[0..len].dup;
		}
	}

	///	Adds the given interval in milliseconds to this time.
	TimeValue opAdd(TimeInterval i) const {
		return TimeValue(normalise(utcMillis + i.value));
	}

	///	Subtracts the given interval in milliseconds from this time.
	TimeValue opSub(TimeInterval i) const {
		return TimeValue(normalise(utcMillis - i.value));
	}

	///	Calculates the interval in milliseconds between this time and another time.
	TimeInterval opSub(TimeValue d) const {
		return cast(TimeInterval) (utcMillis - d.utcMillis
		  + (d.localDayOffset - localDayOffset) * TimeInterval.DAY.value);
	}

	///	Adds the given interval in milliseconds to this time.
	TimeValue opAddAssign(TimeInterval i) {
		utcMillis = normalise(utcMillis + i.value);
		return this;
	}

	///	Subtracts the given interval in milliseconds from this time.
	TimeValue opSubAssign(TimeInterval i) {
		utcMillis = normalise(utcMillis - i.value);
		return this;
	}

	/*	Normalises the time value to the range [0..DAY).
	 *
	 *	Params:
	 *		millis = a millisecond time value that may be out of range.
	 *	Returns:
	 *		the normalised time value.
	 */
	private static int normalise(int millis) {
		if (millis < 0) return millis + TimeInterval.DAY.value;
		if (millis >= TimeInterval.DAY.value) return millis - TimeInterval.DAY.value;
		return millis;
	}

	unittest {
		import std.stdio;

		writefln("Unittest commenced at %s", TimeValue.now.toString);
		assert (TimeValue(19, 45, 32).toString == "19:45:32.000");
		assert (TimeValue(19, 45, 32.487).toString == "19:45:32.487");

		TimeValue t = TimeValue(19, 45, 32, 889);
		assert (t.toString == "19:45:32.889");

		t += TimeInterval.HOUR + 16 * TimeInterval.MINUTE - 25 * TimeInterval.SECOND;
		assert (t.toString == "21:01:07.889");

		assert (TimeValue(23, 45) - TimeValue(22, 45) == TimeInterval.HOUR);
		assert (TimeValue(1, 15) - TimeValue(0, 15) == TimeInterval.HOUR);
		assert (TimeValue(0, 30) - TimeValue(23, 30) == -23 * TimeInterval.HOUR);
	}
}


/**	The DateTime structure represents a combination of date and time.
 *
 *	Bugs: checking for invalid/null components not fully implemented.
 */
struct DateTime {
	///	The date component in UTC.
	DateValue utcDate;
	/// The time component.
	TimeValue utcTime;

	/**	Constructs a DateTime from the given date and time.
	 *	
	 *	Params:
	 *		d = the date component in local time.
	 *		t = the time component.
	 */
	this(DateValue d, TimeValue t = TimeValue(0, 0)) {
		utcDate = d - t.localDayOffset;
		utcTime = t;
	}

	///	Constructs a DateTime representing the given local date and time combination.
	this(int year, int month, int day,
		  int hour, int minute, int second = 0, int ms = 0) {
		this(DateValue(year, month, day), TimeValue(hour, minute, second, ms));
	}

	///	ditto
	this(int year, int month, int day,
		  int hour, int minute, float second) {
		this(DateValue(year, month, day), TimeValue(hour, minute, second));
	}

	version (Windows) {
		/**	Retrieves the current system date and time.
		 *
		 *	Todo: Implement on platforms other than Windows.
		 */
		static @property DateTime now() {
			SYSTEMTIME st;
			DateTime dt;

			GetSystemTime(&st);
			dt.utcDate = DateValue(st.wYear, st.wMonth, st.wDay);
			dt.utcTime = TimeValue(st.wHour * TimeInterval.HOUR.value
			  + st.wMinute * TimeInterval.MINUTE.value + st.wSecond * TimeInterval.SECOND.value
			  + st.wMilliseconds);
			return dt;
		}
	}

	/** Formats this date according to a _format string.
	 *
	 *	Params:
	 *		formatString = the _format string, as defined in smjg.libs.util.datetimeformat.
	 *	Returns:
	 *		the formatted date string.
	 *	Throws:
	 *		DateTimeFormatException if the formatting fails, e.g. because of an error in the
	 *	    _format string.
	 */
	string format(string formatString) const {
		DateTimeComponents dtc = DateTimeComponents(this);
		return formatDateTime(&dtc, dayOfWeek, formatString);
	}

	/**	The date/time as a Unix timestamp, i.e. seconds since the epoch, 00:00 UTC on 1 January
	 *	1970.
	 */
	@property real unixTime() const {
		return (cast(real) (cast(long) utcDate.dayNo - UNIX_EPOCH) * TimeInterval.DAY.value
		  + utcTime.utcMillis) / TimeInterval.SECOND.value;
	}
	
	/// ditto
	@property real unixTime(real ut)
	out (result) {
		real error = (result - unixTime);// * TimeInterval.SECOND.value;
		assert (error > -1);
		assert (error < 1);
	} body {
		long datePart = cast(long) floor(ut / (TimeInterval.DAY / TimeInterval.SECOND));
		utcDate.dayNo = cast(int) (datePart + UNIX_EPOCH);
		utcTime.utcMillis = cast(int)
		  (ut * TimeInterval.SECOND.value - datePart * TimeInterval.DAY.value);
		return ut;
	}

	///	The date component in local time.
	@property DateValue date() const {
		if (utcDate == utcDate.init) {
			throw new DateTimeException("No date to extract");
		} else if (utcTime == utcTime.init) {
			return utcDate;
		} else {
			return utcDate + utcTime.localDayOffset;
		}
	}

	/// ditto
	@property DateValue date(DateValue d)
	out {
		assert (date == d);
	}
	body {
		if (utcTime == utcTime.init) {
			utcDate = d;
		} else {
			utcDate = d - utcTime.localDayOffset;
		}
		return d;
	}

	///	The time component.  Setting this property preserves the local time date.
	@property TimeValue time() const {
		return utcTime;
	}

	/// ditto
	@property TimeValue time(TimeValue t)
	out {
		assert (time == t);
	}
	body {
		setTime(t.utcMillis);
		return t;
	}
	///
	@property int year() const { return date.year; }
	///
	@property MONTH month() const { return date.month; }
	///
	@property int day() const { return date.day; }
	///
	@property DAY_OF_WEEK dayOfWeek() const { return date.dayOfWeek; }
	///
	@property int hour() const { return utcTime.hour; }
	///
	@property int minute() const { return utcTime.minute; }
	///
	@property int second() const { return utcTime.second; }
	///
	@property int millisecond() const { return utcTime.millisecond; }

	///	Returns a string representation in the form e.g. "Sat 8 May 2004 22:12:31.278".
	@property string toString() const {
		return date.toString ~ " " ~ utcTime.toString;
	}

	int opCmp(DateTime dt) const {
		if (date < dt.date) return -1;
		if (date > dt.date) return 1;
		return time - dt.time;
	}

	///	Adds the given date interval to the date component.
	DateTime opAdd(DateInterval di) const {
		return DateTime(date + di, utcTime);
	}

	///	Adds the given time interval to this date and time.
	DateTime opAdd(TimeInterval ti) const {
		DateTime result = this;
		result.setTime(utcTime.utcMillis + ti.value);
		return result;
	}

	///	Adds the given date and time interval to this date and time.
	DateTime opAdd(DateTimeInterval dti) const {
		DateTime result = this;
		result.utcDate += dti.datePart;
		result.setTime(utcTime.utcMillis + dti.timePart.value);
		return result;
	}

	///	Subtracts the given date interval from the date component.
	DateTime opSub(DateInterval di) const {
		return DateTime(date - di, utcTime);
	}

	///	Subtracts the given time interval from this date and time.
	DateTime opSub(TimeInterval ti) const {
		DateTime result = this;
		result.setTime(utcTime.utcMillis - ti);
		return result;
	}

	///	Subtracts the given date and time interval from this date and time.
	DateTime opSub(DateTimeInterval dti) const {
		DateTime result = this;
		result.utcDate -= dti.datePart;
		result.setTime(utcTime.utcMillis - dti.timePart);
		return result;
	}

	///	Determines the time interval between this and another DateTime.
	DateTimeInterval opSub(DateTime dt) const {
		return DateTimeInterval(utcDate - dt.utcDate,
		  utcTime.utcMillis - dt.utcTime.utcMillis);
	}

	///	Adds the given date interval to the date component.
	DateTime opAddAssign(DateInterval di) {
		utcDate += di;
		return this;
	}

	///	Adds the given time interval to this date and time.
	DateTime opAddAssign(TimeInterval ti) {
		setTime(utcTime.utcMillis + ti.value);
		return this;
	}

	///	Adds the given date and time interval to this date and time.
	DateTime opAddAssign(DateTimeInterval dti) {
		utcDate += dti.datePart;
		setTime(utcTime.utcMillis + dti.timePart.value);
		return this;
	}

	///	Subtracts the given date interval from the date component.
	DateTime opSubAssign(DateInterval di) {
		utcDate -= di;
		return this;
	}

	///	Subtracts the given time interval from this date and time.
	DateTime opSubAssign(TimeInterval ti) {
		setTime(utcTime.utcMillis - ti);
		return this;
	}

	///	Subtracts the given date and time interval from this date and time.
	DateTime opSubAssign(DateTimeInterval dti) {
		utcDate -= dti.datePart;
		setTime(utcTime.utcMillis - dti.timePart);
		return this;
	}

	/*	Sets the time within this DateTime object, adjusting the date as
	 *	necessary.
	 */
	private void setTime(int timeValue) {
		if (utcTime.utcMillis == int.min) {
			utcDate += (timeValue - int.min) / TimeInterval.DAY.value;
		} else if (timeValue < 0) {
			utcDate += timeValue / TimeInterval.DAY.value - 1;
			utcTime.utcMillis = timeValue % TimeInterval.DAY.value + TimeInterval.DAY.value;
		} else if (timeValue >= TimeInterval.DAY.value) {
			utcDate += timeValue / TimeInterval.DAY.value;
			utcTime.utcMillis = timeValue % TimeInterval.DAY.value;
		} else {
			utcTime.utcMillis = timeValue;
		}
	}

	unittest {
		import std.stdio;
		
		writefln("Unittest commenced at %s", DateTime.now.toString);
		DateTime dt = DateTime(1979, 9, 24, 3, 5, 20, 123);
		DateValue d = DateValue(1979, 9, 24);
		TimeValue t = TimeValue(3, 5, 20, 123);

		assert (dt == DateTime(d, t));
		assert (dt.date == d);
		assert (dt.time == t);
		assert (dt.toString == "Mon 24 Sep 1979 03:05:20.123");

		DateInterval di = DateInterval(5);
		TimeInterval ti = 21 * TimeInterval.HOUR + 32 * TimeInterval.MINUTE;
		DateTimeInterval dti = DateTimeInterval(di, ti);

		assert ((dt + di).toString == "Sat 29 Sep 1979 03:05:20.123");
		assert ((dt + ti).toString == "Tue 25 Sep 1979 00:37:20.123");
		assert ((dt + dti).toString == "Sun 30 Sep 1979 00:37:20.123");

		dt += dti;
		assert (dt.toString == "Sun 30 Sep 1979 00:37:20.123");
		assert ((dt + ti).toString == "Sun 30 Sep 1979 22:09:20.123");

		dt.utcDate = DateValue(1970, 1, 1);
		dt.utcTime.utcMillis = 0;
		assert (dt.unixTime == 0);

		// set to UTC for following test
		uint tz = timeZoneOffset;
		timeZoneOffset = 0;
		scope(exit) timeZoneOffset = tz;

		// to show that this lib is Y2.038K compliant
		// see http://en.wikipedia.org/wiki/Year_2038_problem
		dt.unixTime = int.max;

		assert (dt.toString == "Tue 19 Jan 2038 03:14:07.000");
		dt.unixTime = cast(long) int.max + 1;
		assert (dt.toString == "Tue 19 Jan 2038 03:14:08.000");
		dt.unixTime = int.min;
		assert (dt.toString == "Fri 13 Dec 1901 20:45:52.000");
		dt.unixTime = cast(long) int.min - 1;
		assert (dt.toString == "Fri 13 Dec 1901 20:45:51.000");
	}
}


/**	Exception thrown when a date/time operation fails, e.g. because some component values are out
 *	of range.
 */
class DateTimeException : Exception {
	this(string msg) { super(msg); }
}

///	The DateTimeComponents structure represents a date and time in structured form.
struct DateTimeComponents {
	/**	Components of the date and time.
	 *	
	 *	$(P Any of these may be set to the .min value of its type, representing an unspecified
	 *	component.)
	 */
	int year          = int.min;
	MONTH month       = MONTH.UNSPEC; /// ditto
	byte day          = byte.min;     /// ditto
	byte hour         = byte.min;     /// ditto
	byte minute       = byte.min;     /// ditto
	byte second       = byte.min;     /// ditto
	short millisecond = short.min;    /// ditto

	deprecated alias millisecond ms;

	invariant() {
		assert (month == byte.min || (month >= 1 && month <= 12));
		/*	don't validate to specific month, as an error could then occur in
		 *	the middle of setting the date
		 */
		assert (day == byte.min || (day >= 1 && day <= 31));
		assert (hour == byte.min || (hour >= 0 && hour < 24));
		assert (minute == byte.min || (minute >= 0 && minute < 60));
		assert (second == byte.min || (second >= 0 && second <= 60));
		assert (millisecond == short.min
		  || (millisecond >= 0 && millisecond < 1000));
	}

	/**	Constructs a DateTimeComponents of the given date, with the time components left
	 *	unspecified.
	 */
	this(DateValue dv) {
		dateValue = dv;
	}

	/**	Constructs a DateTimeComponents of the given time, with the date components left
	 *	unspecified.
	 */
	this(TimeValue tv) {
		timeValue = tv;
	}

	///	Constructs a DateTimeComponents of the given date and time.
	this(DateTime dt) {
		dateValue = dt.date;
		timeValue = dt.time;
	}

	/**	Formats this date and time according to a _format string.
	 *
	 *	Params:
	 *		formatString = the _format string, as defined in smjg.libs.util.datetimeformat.
	 *	Returns:
	 *		the formatted date string.
	 *	Throws:
	 *		DateTimeFormatException  if the formatting fails, e.g. because of an error in the
	 *		                         _format string.
	 */
	string format(string formatString) const {
		return formatDateTime(&this, formatString);
	}

	/**	The DateValue corresponding to the date stored in this DateTimeComponents.
	 *
	 *	Throws:
	 *		DateTimeException  if the date stored in this DateTimeComponents is invalid or
	 *		                   incomplete.
	 */
	@property DateValue dateValue() const {
		return DateValue(year, month, day);
	}

	/// ditto
	@property DateValue dateValue(DateValue dv)
	out {
		assert (dateValue == dv);
	}
	body {
		int d;

		year = dv.extractYear(d);
		month = extractMonthDay(d, isLeapYear(year));
		day = cast(byte) d;

		return dv;
	}

	/**	The TimeValue corresponding to the time stored in this DateTimeComponents.
	 *
	 *	Throws:
	 *		DateTimeException  if the time stored in this DateTimeComponents is invalid or
	 *		                   incomplete.
	 */
	@property TimeValue timeValue() const {
		return TimeValue(hour, minute, second, millisecond);
	}

	/// ditto
	@property TimeValue timeValue(TimeValue tv)
	out {
		assert (timeValue == tv);
	}
	body {
		hour = cast(byte) tv.hour;
		minute = cast(byte) tv.minute;
		second = cast(byte) tv.second;
		millisecond = cast(short) tv.millisecond;

		return tv;
	}
}
