﻿/* ******************************************************************* *\
*                            datetimeformat.d                           *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*                    Workings of date/time formatting                   *
>***********************************************************************<
* Copyright (C) 2005-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
The smjg.libs.util.datetimeformat module implements the generation of formatted date and time
strings.

Usually, one will not use this module directly, but instead use the corresponding properties of
the types defined in smjg.libs.util.datetime.

Format_specifiers:
The letters of the basic Latin alphabet are reserved for format specifiers.  Each specifier is a
letter, or two or more of the same letter consecutively. The letter denotes what piece of
information to format, and the capitalisation and length denote how to format it.  Where
alternative capitalisations are given for a single format, they denote the output capitalisation
of the corresponding datum.
$(TABLE
	$(THEAD
		$(TR
			$(TH Letter)
			$(TH Datum)
			$(TH Specifier)
			$(TH Format)
		)
	)
	$(TBODY
		$(TR
			$(TDR 4,Y)
			$(TDR 4,Year)
			$(TD $(D yy))
			$(TD Two digits)
		)
		$(TR
			$(TD $(D yyy))
			$(TD Exact required length)
		)
		$(TR
			$(TD $(D yyyy))
			$(TD Four digits (longer if necessary))
		)
		$(TR
			$(TD $(D YYY))
			$(TD Astronomical notation (BC represented as negative numbers))
		)
	)
	$(TBODY
		$(TR
			$(TDR 4,B)
			$(TDR 4,BC/AD)
			$(TD $(D B) or $(D b))
			$(TD BC if necessary)
		)
		$(TR
			$(TD $(D BB) or $(D bb))
			$(TD BC or AD)
		)
		$(TR
			$(TD $(D BBB) or $(D bbb))
			$(TD BCE or CE)
		)
		$(TR
			$(TD $(D BBBB) or $(D bbbb))
			$(TD BCE if necessary)
		)
	)
	$(TBODY
		$(TR
			$(TDR 4,M)
			$(TDR 4,Month)
			$(TD $(D m))
			$(TD numeric without leading zero)
		)
		$(TR
			$(TD $(D mm))
			$(TD numeric with leading zero)
		)
		$(TR
			$(TD $(D MMM), $(D Mmm) or $(D mmm))
			$(TD abbreviated name)
		)
		$(TR
			$(TD $(D MMMM), $(D Mmmm) or $(D mmmm))
			$(TD full name)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,D)
			$(TDR 2,Day of month)
			$(TD $(D d))
			$(TD no leading zero)
		)
		$(TR
			$(TD $(D dd))
			$(TD leading zero)
		)
	)
	$(TBODY
		$(TR
			$(TD T)
			$(TD Ordinal suffix)
			$(TD $(D T) or $(D t))
			$(TD Ordinal suffix (TH, ND or RD) of last formatted datum)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,W)
			$(TDR 2,Day of week)
			$(TD $(D WWW), $(D Www) or $(D www))
			$(TD abbreviated name)
		)
		$(TR
			$(TD $(D WWWW), $(D Wwww) or $(D wwww))
			$(TD full name)
		)
	)
	$(TBODY
		$(TR
			$(TDR 4,H)
			$(TDR 4,Hour)
			$(TD $(D H))
			$(TD 24-hour without leading zero)
		)
		$(TR
			$(TD $(D HH))
			$(TD 24-hour with leading zero)
		)
		$(TR
			$(TD $(D h))
			$(TD 12-hour without leading zero)
		)
		$(TR
			$(TD $(D hh))
			$(TD 12-hour with leading zero)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,A)
			$(TDR 2,AM/PM)
			$(TD $(D A) or $(D a))
			$(TD A or P)
		)
		$(TR
			$(TD $(D AA) or $(D aa))
			$(TD AM or PM)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,I)
			$(TDR 2,Minute)
			$(TD $(D i))
			$(TD no leading zero)
		)
		$(TR
			$(TD $(D ii))
			$(TD leading zero)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,S)
			$(TDR 2,Second)
			$(TD $(D s))
			$(TD no leading zero)
		)
		$(TR
			$(TD $(D ss))
			$(TD leading zero)
		)
	)
	$(TBODY
		$(TR
			$(TDR 5,F)
			$(TDR 5,Fraction of a second)
			$(TD $(D f))
			$(TD deciseconds)
		)
		$(TR
			$(TD $(D FF))
			$(TD centiseconds with leading zero)
		)
		$(TR
			$(TD $(D ff))
			$(TD centiseconds without leading zero)
		)
		$(TR
			$(TD $(D FFF))
			$(TD milliseconds with leading zero)
		)
		$(TR
			$(TD $(D fff))
			$(TD milliseconds without leading zero)
		)
	)
	$(TBODY
		$(TR
			$(TDR 2,Z)
			$(TDR 2,Time zone)
			$(TD $(D zzzz))
			$(TD offset from UTC)
		)
	)
)

Format_specifier_examples:
$(TABLE
	$(THEAD
		$(TR
			$(TH Format string)
			$(TH Sample formatted date/time)
		)
	)
	$(TBODY
		$(TR
			$(TD $(D dd/mm/yy))
			$(TD $(D 08/09/05))
		)
		$(TR
			$(TD $(D Www dt Mmm yyyy BB))
			$(TD $(D Thu 8th Sep 2005 AD))
		)
		$(TR
			$(TD $(D h:ii AA))
			$(TD $(D 4:51 PM))
		)
		$(TR
			$(TD $(D yyyy-mm-dd HH:ii:ss zzzz))
			$(TD $(D 2005-09-08 16:51:09 +0100))
		)
	)
)

Literals:
$(P Any non-alphabetic character that doesn't have a special meaning is a literal, i.e. it will be
placed as is in the generated string.  You can literalise a single character by placing a
backquote ($(D `)) immediately before it, or any number of consecutive characters by enclosing
them in single quotes ($(D '...')).)

$(P All letters of the basic Latin alphabet, even those that are not defined as format specifiers,
must be literalised if they are to occur as are in the generated string.  This is to prevent
accidental use of letters that may become format specifiers in future versions of the notation.)

$(P The following characters will always be literals: $(D- : , . /) space and all Unicode codepoints
beyond U+007F.)
Alignment_fields:
$(P (Since 0.5))

$(P Square brackets ($(D [ ])) can be used to create fields.  A field is a piece of formatting that
is padded to a specific width.  Inside a pair of square brackets, the input consists of up to
three parts: the left padding, content and right padding.&nbsp; Either padding part consists of
the same literal character one or more times in succession.  The total number of padding
characters determines the width of the field.  The field alignment is determined by which paddings
are present:)
$(UL
	$(LI If there is only right padding, it will be left aligned.)
	$(LI If there is only left padding, it will be right aligned.)
	$(LI
		If there is padding on both sides, it will be centred.  If the content cannot be centred
		exactly, then the output will have one more padding character on the left, unless in the
		format string the right padding is longer than the left padding, in which case there will
		be one more padding character on the right.)
)
$(P An alternative is to place a number just inside the square bracket, next to a single padding
character.  Then the number is the field width.  For left-aligned and right-aligned fields, the
number must be on the same side as the padding character.&nbsp; For centred fields, the padding
character is given on each side, and the side on which the number is placed will receive the extra
padding character if the content cannot be centred exactly.)

$(P In the unlikely event that you may wish to use a digit as a padding character, it must be
literalised.)
---
[------Wwww.....]
---
$(P This is a centred, eleven-character field containing the name of the day of the week. The
generated string will be one of these:)
---
---Sunday..
---Monday..
--Tuesday..
-Wednesday.
--Thursday.
---Friday..
--Saturday.
---
$(P The format string $(D[11-Wwww.]) has the same meaning.)

$(P The content can contain any number of format specifiers and literals.  For example:)
---
[d/m/yyy           ]HH:ii:ss
---
$(P creates, if outputting a list of date/time values, a column for the date and a column for the
time.  Sample output:)
---
24/9/1979  03:05:42
15/11/1983 21:43:05
3/4/991    13:57:24
---

Collapsible_portions:
$(P Curly brackets ($(D { })) can be used to create a collapsible portion.  This is a portion that
will appear in the output only if at least one of the format specifiers within it actually
generates something.&nbsp; Under the current spec, the only conditions in which a format specifier
may generate nothing are:)
$(UL
	$(LI The $(D B), $(D b), $(D BBBB) or $(D bbbb) specifier is used, and the date is AD.)
	$(LI
		The requested component (or a component required to calculate what is requested) is
		unspecified in the date/time object being formatted.)
	$(LI The time zone is not output for any date/time object in which the hour is unspecified.)
)
---
d Mmm yyy{ B}{ HH:ii:ss zzzz}
---
$(P Formats the date, and the time if present, with only one space between them
whether the date is BC or AD.)
---
{d }{Mmm }yyy BB
---
$(P For applications in which an exact date, the month and year or just the
year may be known, to prevent leading spaces.)
---
HH:ii{:ss}{.FFF}
---
$(P For applications in which a time is always known to the precision of
minutes, but may sometimes be known more precisely.)

See_Also:
	$(LINK2 http://pr.stewartsplace.org.uk/d/sutil/datetime_format.html,the original proposal)
	(includes discussion)
Since: 0.02
Todo:  implement timezone TLAs.
 */
module smjg.libs.util.datetimeformat;

private import
	smjg.libs.util.datetime,
	std.ascii,
	std.utf;
private import std.string : toStringz;
private import std.conv   : to;


/**	Formats dt according to formatString.
 *
 *	Returns:
 *		the formatted date string.
 *	Throws:
 *		DateTimeFormatException  if the formatting fails, e.g. because of an error in the format
 *		                         string.
 *		UtfException             if formatString is not a correctly-formed UTF-8 string.
 */
string formatDateTime(const(DateTimeComponents)* dt, string formatString) {
	std.utf.validate(formatString);
	try {
		return formatDateTime(dt, dt.dateValue.dayOfWeek, formatString);
	} catch (DateTimeException e) {
		//debug writefln("Incomplete date");
		return formatDateTime(dt, DAY_OF_WEEK.init, formatString);
	}
}


package string formatDateTime(const(DateTimeComponents)* dt, DAY_OF_WEEK dayOfWeek,
                              string formatString) {
	std.utf.validate(formatString);
	bool nonNull;
	immutable(char)* charPos = toStringz(formatString);
	scope(success) assert (*charPos == '\0');

	return formatDateTime(dt, dayOfWeek, charPos, nonNull, '\0');
}


private {
	//alias smjg.libs.d12comp.toString toString;

	// taken from Phobos (where it is private in D2)
	const ubyte[256] UTF8stride = [
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
		0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
		0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
		0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
		2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
		2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
		3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
		4,4,4,4,4,4,4,4,5,5,5,5,6,6,0xFF,0xFF,
	];

	string formatDateTime(const(DateTimeComponents)* dt, DAY_OF_WEEK dayOfWeek,
		  ref immutable(char)* charPos, out bool nonNull, char portionEnd) {
		// function uses null-terminated string to make finding the end easier
		int lastNumber = int.min;
		string result;

		while (*charPos != portionEnd) {
			if (beginsElement(*charPos)) {
				bool newNonNull;
				result ~= formatElement(dt, dayOfWeek, charPos, newNonNull, lastNumber);
				if (newNonNull) nonNull = true;
			} else if (beginsLiteral(*charPos)) {
				result ~= formatLiteral(charPos);
			} else switch (*charPos) {
				case '\0':		// unclosed portion
					assert (portionEnd == '}');
					throw new DateTimeFormatException(E_UNCLOSED_COLLAPSIBLE);

				case '}':
					throw new DateTimeFormatException(E_UNOPENED_COLLAPSIBLE);

				case ']':
					throw new DateTimeFormatException(E_UNOPENED_FIELD);

				default: // self-literal character
					result ~= *(charPos++);
			}
		}

		return result;
	}


	/*	Processes a single format element.  A format element is any of the following:
	 *	- an alphabetical format specifier
	 *	- a collapsible portion
	 *	- an alignment field
	 *	Literals and alignment field widths are not included.  The purpose is
	 *	to deal with those elements that cannot be part of an alignment field
	 *	padding or width.
	 */
	string formatElement(const(DateTimeComponents)* dt, DAY_OF_WEEK dayOfWeek,
	                     ref immutable(char)* charPos, out bool nonNull, ref int lastNumber)
	in {
		assert (beginsElement(*charPos));
	} body {
		switch (*charPos) {
			case '[': {
				charPos++;
				string portion = formatField(dt, dayOfWeek, charPos, nonNull);
				charPos++;
				return portion;
			}

			case '{': {
				charPos++;
				string portion = formatDateTime(dt, dayOfWeek, charPos, nonNull, '}');
				charPos++;
				return nonNull ? portion : null;
			}

			default:
				char letter = cast(char) toLower(*charPos);
				immutable(char)* beginSpec = charPos;

				do {
					++charPos;
				} while (toLower(*charPos) == letter);

				string formatted = formatBySpec(dt, dayOfWeek,
				  beginSpec[0 .. charPos-beginSpec], lastNumber);

				if (formatted.length != 0) {
					nonNull = true;
					return formatted;
				} else {
					return null;
				}
		}
	}


	string formatLiteral(ref immutable(char)* charPos)
	in {
		assert (beginsLiteral(*charPos));
	} body {
		switch (*charPos) {
			case '`': {		// literal character
				if (*++charPos == '\0') {
					throw new DateTimeFormatException(E_MISSING_LITERAL);
				}
				uint len = UTF8stride[*charPos];
				scope(exit) charPos += len;
				return charPos[0..len];
			}

			case '\'': {	// literal portion
				immutable(char)* beginLiteral = ++charPos;
				while (*charPos != '\'') {
					if (*charPos == '\0') {
						throw new DateTimeFormatException(E_UNCLOSED_LITERAL);
					}
					charPos++;
				}
				return beginLiteral[0 .. (charPos++) - beginLiteral];
			}

			default: assert (false);
		}
	}


	struct Piece {
		dstring dtext;
		string  text;
		ubyte type;  // 0 = raw, 1 = literal; 2 = formatted

		@property uint asNumber() {
			if (dtext.length > 9) throw new DateTimeFormatException(E_OVERFLOW_WIDTH);
			uint result;
			foreach (c; dtext) {
				assert (c >= '0' && c <= '9');
				result = result * 10 + (c - '0');
			}
			return result;
		}
	}


	string formatField(const(DateTimeComponents)* dt, DAY_OF_WEEK dayOfWeek,
	                   ref immutable(char)* charPos, out bool nonNull) {
		Piece[] pieces;

		// first parse the format string within the [...]
		{
			Piece[] tempPieces;
			int lastNumber = int.min;

			while (*charPos != ']') {
				if (beginsElement(*charPos)) {
					bool newNonNull;
					tempPieces ~= Piece(null, formatElement(dt, dayOfWeek, charPos, newNonNull, lastNumber), 2);
					if (newNonNull) nonNull = true;
				} else if (beginsLiteral(*charPos)) {
					tempPieces ~= Piece(null, formatLiteral(charPos), 1);
				} else switch (*charPos) {
					case '\0':
						throw new DateTimeFormatException(E_UNCLOSED_FIELD);

					case '}':
						throw new DateTimeFormatException(E_UNOPENED_COLLAPSIBLE);

					default: {
						immutable(char)* begin = charPos;
						do {
							charPos++;
						} while (*charPos != '\0' && *charPos != ']' && *charPos != '}'
						  && !beginsElement(*charPos) && !beginsLiteral(*charPos));

						tempPieces ~= Piece(null, begin[0 .. charPos - begin], 0);
					}
				}
			}

			/*	convert tempPieces into a form in which
			 *	- no two consecutive tempPieces have the same type
			 *	- only non-literalised numbers have type 0
			 */
			ubyte lastType = ubyte.max;

			foreach (piece; tempPieces) {
				switch (piece.type) {
				case 0:
					foreach (dchar c; piece.text) {
						if (isDigit(c)) {
							if (lastType == 0) {
								pieces[$-1].dtext ~= c;
							} else {
								pieces ~= Piece([c], null, 0);
								lastType = 0;
							}
						} else {
							if (lastType == 1) {
								pieces[$-1].dtext ~= c;
							} else {
								pieces ~= Piece([c], null, 1);
								lastType = 1;
							}
						}
					}
					break;

				case 1:
					if (lastType == 1) {
						pieces[$-1].dtext ~= toUTF32(piece.text);
					} else {
						pieces ~= Piece(toUTF32(piece.text), null, 1);
						lastType = 1;
					}
					break;

				case 2:
					if (lastType == 2) {
						pieces[$-1].text ~= piece.text;
					} else {
						pieces ~= piece;
						lastType = 2;
					}
					break;

				default:
					assert (false);
				}
			}
		}

		if (pieces.length < 2) throw new DateTimeFormatException(E_INCOMPLETE_FIELD);

		// detect the field width/padding
		dchar padLeft, padRight;
		size_t fieldWidth = 0;
		bool moreOnRight;

		if (pieces[0].type == 0) {
			// field width on left
			if (pieces[$-1].type == 0) throw new DateTimeFormatException(E_DOUBLE_WIDTH);

			fieldWidth = pieces[0].asNumber;
			if (fieldWidth == 0) throw new DateTimeFormatException(E_ZERO_FIELD);
			if (pieces[1].type != 1) throw new DateTimeFormatException(E_INCOMPLETE_FIELD);

			pieces = pieces[1..$];
			padLeft = pieces[0].dtext[0];
			pieces[0].dtext = pieces[0].dtext[1..$];
			if (pieces[$-1].type == 1) {
				padRight = pieces[$-1].dtext[$-1];
				pieces[$-1].dtext.length = pieces[$-1].dtext.length - 1;
			}

		} else if (pieces[$-1].type == 0) {
			// field width on right
			moreOnRight = true;
			fieldWidth = pieces[$-1].asNumber;
			if (fieldWidth == 0) throw new DateTimeFormatException(E_ZERO_FIELD);
			if (pieces[$-2].type != 1) throw new DateTimeFormatException(E_INCOMPLETE_FIELD);

			pieces = pieces[0..$-1];
			padRight = pieces[$-1].dtext[$-1];
			pieces[$-1].dtext.length = pieces[$-1].dtext.length - 1;
			if (pieces[0].type == 1) {
				padLeft = pieces[0].dtext[0];
				pieces[0].dtext = pieces[0].dtext[1..$];
			}

		} else {
			// field width given by number of padding characters
			if (pieces[0].type == 1) {
				padLeft = pieces[0].dtext[0];
				for (fieldWidth = 1;
				  fieldWidth < pieces[0].dtext.length && pieces[0].dtext[fieldWidth] == padLeft;
				  fieldWidth++) {}
				pieces[0].dtext = pieces[0].dtext[fieldWidth..$];
			}
			if (pieces[$-1].type == 1) {
				padRight = pieces[$-1].dtext[$-1];
				uint pos;
				for (pos = pieces[$-1].dtext.length - 1;
				  pos > 0 && pieces[$-1].dtext[pos - 1] == padRight;
				  pos--) {}
				if (pieces[$-1].dtext.length - pos > fieldWidth) moreOnRight = true;
				fieldWidth += pieces[$-1].dtext.length - pos;
				pieces[$-1].dtext.length = pos;
			}
		}

		assert (fieldWidth != 0);

		debug (datetimeformat) {
			writefln("padding chars: %s %s.", padLeft == dchar.init ? "none" : [padLeft],
			  padRight == dchar.init ? "none" : [padRight]);
			writefln("width: %d", fieldWidth);
			writefln("%d pieces", pieces.length);
		}

		// read the field format - now use it
		// but first, concatenate and measure the content
		size_t contentLength;
		string formattedContent;
		foreach (piece; pieces) {
			assert (piece.dtext.length == 0 || piece.text.length == 0);
			if (piece.text.length == 0) {
				formattedContent ~= toUTF8(piece.dtext);
				contentLength += piece.dtext.length;
			} else {
				formattedContent ~= piece.text;
				contentLength += toUCSindex(piece.text, piece.text.length);
			}
		}
		debug (datetimeformat) writefln("content length %d: %s", contentLength, formattedContent);

		if (contentLength > fieldWidth && !allowFieldOverflow) {
			throw new DateTimeFormatException(E_FIELD_OVERFLOW);
		}
		if (contentLength >= fieldWidth) return formattedContent;
		assert (formattedContent.length == toUTFindex(formattedContent, contentLength));

		// distribute padding
		uint padWidth = fieldWidth - contentLength, padLeftWidth = 0, padRightWidth = 0;
		if (padLeft == dchar.init) {
			padRightWidth = padWidth;
		} else if (padRight == dchar.init) {
			padLeftWidth = padWidth;
		} else {
			padLeftWidth = padRightWidth = padWidth / 2;
			if (padWidth % 2 == 1) {
				if (moreOnRight) {
					padRightWidth++;
				} else {
					padLeftWidth++;
				}
			}
		}
		debug (datetimeformat) writefln("Padding distribution: %d %d %d = %d",
		  padLeftWidth, contentLength, padRightWidth, fieldWidth);
		assert (padLeftWidth + contentLength + padRightWidth == fieldWidth);

		// now do it!
		char[] result;

		for (int i = 0; i < padLeftWidth; i++) encode(result, padLeft);
		result ~= formattedContent;
		for (int i = 0; i < padRightWidth; i++) encode(result, padRight);
		return cast(string) result;
	}

	bool beginsElement(char c) { return isAlpha(c) || c == '[' || c == '{'; }
	bool beginsLiteral(char c) { return c == '\'' || c == '`'; }

	const string	DIGITS12 = "110123456789";
					/+TEN = DIGITS12[1..3],
					ELEVEN = DIGITS12[0..2],
					TWELVE = DIGITS12[3..5];+/

	/*const E_BAD_UTF
	  = "Error in date/time format string: invalid UTF-8 sequence";*/
	const E_MISSING_LITERAL
	  = "Error in date/time format string: missing character after '`'";
	const E_UNCLOSED_LITERAL
	  = "Error in date/time format string: unterminated literal portion";
	const E_UNCLOSED_FIELD
	  = "Error in date/time format string: '[' without matching ']'";
	const E_UNCLOSED_COLLAPSIBLE
	  = "Error in date/time format string: '{' without matching '}'";
	const E_UNOPENED_FIELD
	  = "Error in date/time format string: ']' without matching '['";
	const E_UNOPENED_COLLAPSIBLE
	  = "Error in date/time format string: '}' without matching '{'";
	const E_INCOMPLETE_FIELD
	  = "Error in date/time format string: Incomplete alignment field";
	const E_ZERO_FIELD
	  = "Error in date/time format string: Zero-width alignment field";
	const E_DOUBLE_WIDTH
	  = "Error in date/time format string: Width of alignment field doubly specified";
	const E_OVERFLOW_WIDTH
	  = "Error in date/time format string: Field width too large";
	const E_FIELD_OVERFLOW
	  = "Date/time formatting failed: Insufficient field width to hold content";
	const E_BC_YY
	  = "Date/time formatting failed: Format 'yy' for BC dates undefined";
	const E_INVALID_DATE_TIME
	  = "Date/time formatting failed: Invalid date/time";

	//int lastNumber = -1;

	string formatBySpec(const(DateTimeComponents)* dt, DAY_OF_WEEK dow,
		  string spec, ref int lastNumber) {
		with (*dt) switch (spec) {
			/*	no need to validate the date itself, as DateTimeComponents
			 *	invariants do it anyway
			 */
			case "yy":
				lastNumber = year;
				if (year == year.min) return null;
				if (year <= 0) {
					throw new DateTimeFormatException(E_BC_YY);
				}
				return formatTwoDigit(cast(byte) (year % 100));

			case "yyy":
				lastNumber = year;
				return year == year.min ? null :
				  to!string(year > 0 ? year : (lastNumber = 1 - dt.year));

			case "yyyy":
				lastNumber = year;
				return year == year.min ? null :
				  std.string.format("%04d", year > 0 ? year :
				  (lastNumber = 1 - dt.year));

			case "YYY":
				lastNumber = year < 0 ? -year : year; // year.min remains the same
				return year == year.min ? null : to!string(year);

			case "b":
				return (year == year.min || year > 0) ? null : "bc";

			case "bb":
				return year == year.min ? null : year > 0 ? "ad" : "bc";

			case "bbb":
				return year == year.min ? null : year > 0 ? "ce" : "bce";

			case "bbbb":
				return (year == year.min || year > 0) ? null : "bce";

			case "B":
				return (year == year.min || year > 0) ? null : "BC";

			case "BB":
				return year == year.min ? null : year > 0 ? "AD" : "BC";

			case "BBB":
				return year == year.min ? null : year > 0 ? "CE" : "BCE";

			case "BBBB":
				return (year == year.min || year > 0) ? null : "BCE";

			case "m":
				if (month == month.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = month;
				return format12(month);

			case "mm":
				if (month == month.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = month;
				char[] fmt = new char[2];
				if (month < 10) {
					fmt[0] = '0';
					fmt[1] = cast(char) ('0' + month);
				} else {
					fmt[0] = '1';
					fmt[1] = cast(char) ('0' - 10 + month);
				}
				return cast(string) fmt;

			case "mmm":
				return month == month.min ? null : SHORT_L_MONTH_NAME[month];

			case "Mmm":
				return month == month.min ? null : SHORT_MONTH_NAME[month];

			case "MMM":
				return month == month.min ? null : SHORT_U_MONTH_NAME[month];

			case "mmmm":
				return month == month.min ? null : LONG_L_MONTH_NAME[month];

			case "Mmmm":
				return month == month.min ? null : LONG_MONTH_NAME[month];

			case "MMMM":
				return month == month.min ? null : LONG_U_MONTH_NAME[month];

			case "d":
				if (day == day.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = day;
				return to!string(day);

			case "dd":
				if (day == day.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = day;
				return formatTwoDigit(day);

			case "t":
				return ordinalSuffix(lastNumber, false);

			case "T":
				return ordinalSuffix(lastNumber, true);

			case "www":
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return SHORT_L_DAY_NAME[dow];

			case "Www":
				debug (datetimeformat) writefln("Day of week: %d", cast(byte) dow);
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return SHORT_DAY_NAME[dow];

			case "WWW":
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return SHORT_U_DAY_NAME[dow];

			case "wwww":
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return LONG_L_DAY_NAME[dow];

			case "Wwww":
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return LONG_DAY_NAME[dow];

			case "WWWW":
				if (dow == dow.min) {
					lastNumber = int.min;
					return null;
				}
				return LONG_U_DAY_NAME[dow];

			case "h":
				if (hour == hour.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = hour;
				if (hour == 0) {
					return DIGITS12[3..5];
				} else if (hour <= 12) {
					return format12(hour);
				} else {
					return format12(hour - 12);
				}

			case "hh":
				if (hour == hour.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = hour;
				if (hour == 0) {
					return DIGITS12[3..5];
				} else if (hour <= 12) {
					return formatTwoDigit(hour);
				} else {
					return formatTwoDigit(hour - 12);
				}

			case "H":
				if (hour == hour.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = hour;
				return to!string(hour);

			case "HH":
				if (hour == hour.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = hour;
				return formatTwoDigit(hour);

			case "a":
				return hour == hour.min ? null :
				  hour < 12 ? "a" : "p";

			case "aa":
				return hour == hour.min ? null :
				  hour < 12 ? "am" : "pm";

			case "A":
				return hour == hour.min ? null :
				  hour < 12 ? "A" : "P";

			case "AA":
				return hour == hour.min ? null :
				  hour < 12 ? "AM" : "PM";

			case "i":
				if (minute == minute.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = minute;
				return to!string(minute);

			case "ii":
				if (minute == minute.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = minute;
				return formatTwoDigit(minute);

			case "s":
				if (second == second.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = second;
				return to!string(second);

			case "ss":
				if (second == second.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = second;
				return formatTwoDigit(second);

			case "f":
				if (millisecond == millisecond.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = millisecond / 100;
				return DIGITS12[lastNumber+2..lastNumber+3];

			case "ff":
				if (millisecond == millisecond.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = millisecond / 10;
				return to!string(lastNumber);

			case "FF":
				if (millisecond == millisecond.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = millisecond / 10;
				return formatTwoDigit(cast(byte) lastNumber);

			case "fff":
				if (millisecond == millisecond.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = millisecond;
				return to!string(millisecond);

			case "FFF":
				if (millisecond == millisecond.min) {
					lastNumber = int.min;
					return null;
				}
				lastNumber = millisecond;
				return std.string.format("%03d", millisecond);

			case "zzzz":
				return hour == hour.min ? null :
				  timeZoneOffset >= 0 ?
				    std.string.format("+%02d%02d", timeZoneOffset / 60,
				      timeZoneOffset % 60) :
				  std.string.format("-%02d%02d", -timeZoneOffset / 60,
				      -timeZoneOffset % 60);

			default:
				throw new DateTimeFormatException(cast(string)
				  ("Error in date/time format string: Undefined format specifier '" ~ spec ~ "'"));
		}
	}

	string formatTwoDigit(int b)
	in {
		assert (b == byte.min || (b >= 0 && b <= 99));
	} body {
		if (b == byte.min) return null;
		char[] fmt = new char[2];
		fmt[0] = cast(byte) ('0' + b / 10);
		fmt[1] = cast(byte) ('0' + b % 10);
		return cast(string) fmt;
	}

	string format12(int b)
	in {
		assert (b >= 0);
		assert (b <= 12);
	} body {
		switch (b) {
			case 10: return DIGITS12[1..3];
			case 11: return DIGITS12[0..2];
			case 12: return DIGITS12[3..5];
			default: return DIGITS12[2+b .. 3+b];
		}
	}

	string ordinalSuffix(int lastNumber, bool upperCase) {
		if (lastNumber < 0) return null;
		lastNumber %= 100;
		if (lastNumber >= 4 && lastNumber <= 20) {
			return upperCase ? "TH" : "th";
		}
		switch (lastNumber % 10) {
			case 1:
				return upperCase ? "ST" : "st";

			case 2:
				return upperCase ? "ND" : "nd";

			case 3:
				return upperCase ? "RD" : "rd";

			default:
				return upperCase ? "TH" : "th";
		}
	}
}


///	Exception thrown if there was a problem in formatting a date or time.
class DateTimeFormatException : Exception {
	private this(string msg) {
		super(msg);
	}
}


///	Short (three-letter) names of the days of the week.
immutable string[7] SHORT_L_DAY_NAME = [
	DAY_OF_WEEK.SUN: "sun", "mon", "tue", "wed", "thu", "fri", "sat"
];

///	Short (three-letter) names of the days of the week.
immutable string[7] SHORT_U_DAY_NAME = [
	DAY_OF_WEEK.SUN: "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"
];

///	Full names of the days of the week.
immutable string[7] LONG_L_DAY_NAME = [
	DAY_OF_WEEK.SUN: "sunday", "monday", "tuesday", "wednesday",
	  "thursday", "friday", "saturday"
];

///	Full names of the days of the week.
immutable string[7] LONG_U_DAY_NAME = [
	DAY_OF_WEEK.SUN: "SUNDAY", "MONDAY", "TUESDAY", "WEDNESDAY",
	  "THURSDAY", "FRIDAY", "SATURDAY"
];

///	Short (three-letter) names of the months of the year.
immutable string[13] SHORT_L_MONTH_NAME = [
	['\xFF', '\xFF', '\xFF'],
	MONTH.JAN: "jan", "feb", "mar", "apr", "may", "jun",
	"jul", "aug", "sep", "oct", "nov", "dec"
];

///	Short (three-letter) names of the months of the year.
immutable string[13] SHORT_U_MONTH_NAME = [
	['\xFF', '\xFF', '\xFF'],
	MONTH.JAN: "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
	"JUL", "AUG", "SEP", "OCT", "NOV", "DEC"
];

///	Full names of the months of the year.
immutable string[13] LONG_L_MONTH_NAME = [
	null, MONTH.JAN: "january", "february", "march", "april", "may", "june",
	"july", "august", "september", "october", "november", "december"
];

///	Full names of the months of the year.
immutable string[13] LONG_U_MONTH_NAME = [
	null, MONTH.JAN: "JANUARY", "FEBRUARY", "MARCH", "APRIL", "MAY", "JUNE",
	"JULY", "AUGUST", "SEPTEMBER", "OCTOBER", "NOVEMBER", "DECEMBER"
];

unittest {
	import std.stdio;

	writefln("Unittest commenced at %s", DateTime.now.toString);

	uint tz = timeZoneOffset;
	timeZoneOffset = 60;
	scope(exit) timeZoneOffset = tz;

	DateTime dt = DateTime(2005, 9, 8, 16, 51, 9.427);
	// basic formatting
	assert (dt.format("dd/mm/yy") == "08/09/05");
	assert (dt.format("Www dt Mmm yyyy BB") == "Thu 8th Sep 2005 AD");
	assert (dt.format("h:ii AA") == "4:51 PM");
	assert (dt.format("yyyy-mm-dd HH:ii:ss zzzz") == "2005-09-08 16:51:09 +0100");
	assert (dt.format("HH:ii:ss.FFF") == "16:51:09.427");
	// alignment fields
	assert (dt.format("[------Wwww.....]") == "--Thursday.");
	assert (dt.format("[11-Wwww.]") == "--Thursday.");
	assert (dt.format("[-----Wwww......]") == "-Thursday..");
	assert (dt.format("[-Wwww.11]") == "-Thursday..");
	assert (dt.format("[9`1Www]") == "111111Thu");
	assert (dt.format("[`1Wwww-10]") == "1Thursday-");
	assert (dt.format("[d/m/yyy           ]HH:ii:ss") == "8/9/2005   16:51:09");
	// collapsible portions (control)
	assert (dt.format("d Mmm yyy{ B}{ HH:ii:ss zzzz}") == "8 Sep 2005 16:51:09 +0100");
	assert (dt.format("{d }{Mmm }yyy BB") == "8 Sep 2005 AD");
	assert (dt.format("HH:ii{:ss}{.FFF}") == "16:51:09.427");

	// create a DTC to test collapsible portions
	DateTimeComponents dtc = DateTimeComponents(dt);
	assert (dtc.format("HH:ii{:ss}{.FFF}") == "16:51:09.427");
	dtc.millisecond = short.min;
	assert (dtc.format("HH:ii{:ss}{.FFF}") == "16:51:09");
	dtc.second = byte.min;
	assert (dtc.format("HH:ii{:ss}{.FFF}") == "16:51");
	assert (dtc.format("d Mmm yyy{ B}{ HH:ii:ss zzzz}") == "8 Sep 2005 16:51: +0100");
	dtc.hour = byte.min;
	assert (dtc.format("d Mmm yyy{ B}{ HH:ii:ss zzzz}") == "8 Sep 2005 :51: ");
	dtc.minute = byte.min;
	assert (dtc.format("d Mmm yyy{ B}{ HH:ii:ss zzzz}") == "8 Sep 2005");
	assert (dtc.format("{d }{Mmm }yyy BB") == "8 Sep 2005 AD");
	dtc.month = MONTH.UNSPEC;
	assert (dtc.format("{d }{Mmm }yyy BB") == "8 2005 AD");
	dtc.day = byte.min;
	assert (dtc.format("{d }{Mmm }yyy BB") == "2005 AD");

	// nesting of fields and collapsible portions
	dtc = DateTimeComponents(dt);
	assert (dtc.format("[13 Mmmm [d..]]") == " September 8.");
	assert (dtc.format("[13 Mmmm{ d}]") == "  September 8");
	dtc.day = byte.min;
	assert (dtc.format("[13 Mmmm{ d}]") == "    September");
	assert (dtc.format("{[13 Mmmm{ d}]}") == "    September");
	dtc.month = MONTH.UNSPEC;
	assert (dtc.format("{[13 Mmmm{ d}]}") == "");
	dtc.day = 8;
	assert (dtc.format("{[13 Mmmm{ d}]}") == "            8");
}
