﻿/* ******************************************************************* *\
*                                dimval.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*                      Dimensioned value arithmetic                     *
>***********************************************************************<
* Copyright (C) 2012 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
This module is used to perform arithmetic on physical measurements (length, time, etc.) with
compile-time checking that the values combined are of compatible dimensions.

Since:
	0.10
*/
module smjg.libs.util.dimval;

//import std.stdio;
import std.typetuple;

template isDimensionedValue(T : DimensionedValue!(V, /+D1,+/ D), V, /+int D1,+/ D...) {
	enum bool isDimensionedValue = D.length != 0;
}
/// Detects whether T is a DimensionedValue.
template isDimensionedValue(T) {
	enum bool isDimensionedValue = false;
}


template DimensionedValue(Base) {
	alias Base DimensionedValue;
}

/**
The DimensionedValue struct template is used to declare a value that represents a physical
measurement and therefore has dimensions such as length, time, etc.

$(P Dimensions correspond to multiplicatively independent measurements.  Which dimensions
correspond to which measurements is left up to the application.  It is the application's
responsibility to ensure that a given dimension is always measured in the same units.)

$(P Arithmetic and comparison operations are implemented in terms of those of the base types, and
the the base type of the returned value determined accordingly.  They are allowed according to the
following table:)
$(TABLE
	$(THEAD
		$(TR
			$(TH Operator)
			$(TH Allowed dimensions)
			$(TH Returned dimensions)
		)
	)
	$(TBODY
		$(TR
			$(TD $(D +), $(D -), $(D %))
			$(TD Same)
			$(TD Same)
		)
		$(TR
			$(TD $(D *), $(D /))
			$(TD Any)
			$(TD Combined)
		)
		$(TR
			$(TD $(D +=), $(D -=), $(D %=))
			$(TD Same)
			$(TD Same)
		)
		$(TR
			$(TD $(D *=), $(D /=))
			$(TD Dimensionless)
			$(TD Same)
		)
		$(TR
			$(TD $(D ==), $(D !=), $(D <), $(D <=), $(D >), $(D >=))
			$(TD Same)
			$(TD Dimensionless (just a boolean))
		)
	)
)

Params:
	Base = the base type, in which the actual value of the measurement is stored.
	Dims = exponents of zero or more further dimensions (must be of type int).

Preconditions:
	The final given dimension must not be 0.  (Just leave it off.)

Example:
	Suppose the first dimension is length, and the second dimension is time.
---
DimensionedValue!(real, 1)     length;
DimensionedValue!(real, 2)     area;
DimensionedValue!(real, 0,  1) time;
DimensionedValue!(real, 1, -1) speed;
DimensionedValue!(real, 1, -2) acceleration;
---
*/
struct DimensionedValue(Base, /+int Dim1,+/ Dims...)
		if (Dims.length != 0 && !isDimensionedValue!(Base) && CheckDim!(Dims)) {
	alias Base BaseType;
	alias Dims Dimensions;
	/// The _value of the measurement.
	Base value;

	/// Constructs a DimensionedValue that holds the value v.
	this()(in BaseType v) {
		value = v;
	}

	/** Constructs a DimensionedValue from another of the same dimensions and a base type
	 *	implicitly convertible to the base type of this one.
	 */
	this(T : DimensionedValue!(U, Dims), U)(in T v) {
		value = v.value;
	}

	bool opEquals(T : DimensionedValue!(U, Dims), U)(in T v) const {
		return value == v.value;
	}

	int opCmp(T : DimensionedValue!(U, Dims), U)(in T v) const {
		static if (is(U == BaseType)) {
			return typeid(BaseType).compare(&this.value, &v.value);
		} else {
			if (value < v.value) return -1;
			if (value > v.value) return 1;
			return 0;
		}
	}

	hash_t toHash() const {
		return typeid(BaseType).getHash(&value);
	}

	/**	The type of a DimensionedValue of this type multiplied by a T.
	 *
	 *	Bugs:
	 *		As of DMD 2.058, it doesn't work through an alias of the DimensionedValue type
	 *		($(BZ 7447))
	 */
	template MulType(T) {
		static if (isDimensionedValue!(T)) {
			alias DimensionedValue!(MulValueType!(T), NormDim!(DimOps!(Dims).mul!(T.Dimensions)))
				MulType;
		} else {
			alias DimensionedValue!(MulValueType!(T), DimOps!(Dims).id) MulType;
		}
	}

	/**	The type of a DimensionedValue of this type divided by a T.
	 *
	 *	Bugs:
	 *		As of DMD 2.058, it doesn't work through an alias of the DimensionedValue type
	 *		($(BZ 7447))
	 */
	template DivType(T) {
		static if (isDimensionedValue!(T)) {
			alias DimensionedValue!(DivValueType!(T), NormDim!(DimOps!(Dims).div!(T.Dimensions)))
				DivType;
		} else {
			alias DimensionedValue!(DivValueType!(T), DimOps!(Dims).id) DivType;
		}
	}

	/// The reciprocal type to this DimensionedValue.
	alias DimensionedValue!(BaseType, DimOps!(Dims).inv) InvType;

	/// The type of a T divided by a DimensionedValue of this type.
	template DivRType(T) {
		static if (isDimensionedValue!(T)) {
			alias T.DivType!(DimensionedValue) DivRType;
		} else {
			alias DimensionedValue!(typeof(T.init / value), DimOps!(Dims).inv) DivRType;
		}
	}

	auto opBinary(string op, T : DimensionedValue!(U, Dims), U)(in T v) const
			if (op == "+" || op == "-" || op == "%") {
		// don't use auto here as constancy gets in the way
		typeof(mixin("BaseType.init " ~ op ~ "v.value.init"))
			newValue = mixin("value " ~ op ~ " v.value");
		return DimensionedValue!(typeof(newValue), Dims)(newValue);
	}

	MulType!(T) opBinary(string op : "*", T)(in T v) const
			if (isDimensionedValue!(T)) {
		static if (isDimensionedValue!(MulType!(T))) {
			return MulType!(T)(value * v.value);
		} else {
			return value * v.value;
		}
	}
	MulType!(T) opBinary(string op : "*", T)(in T v) const
			if (!isDimensionedValue!(T)) {
		return MulType!(T)(value * v);
	}

	MulType!(T) opBinaryRight(string op : "*", T)(in T v) const
			if (!isDimensionedValue!(T)) {
		return MulType!(T)(v * value);
	}

	DivType!(T) opBinary(string op : "/", T)(in T v) const
			if (isDimensionedValue!(T)) {
		static if (isDimensionedValue!(DivType!(T))) {
			return DivType!(T)(value / v.value);
		} else {
			return value / v.value;
		}
	}
	DivType!(T) opBinary(string op : "/", T)(in T v) const
			if (!isDimensionedValue!(T)) {
		return DivType!(T)(value / v);
	}

	DivRType!(T) opBinaryRight(string op : "/", T)(in T v) const
			if (!isDimensionedValue!(T)) {
		return DivRType!(T)(v / value);
	}

	DimensionedValue opOpAssign(string op, T : DimensionedValue!(U, Dims), U)(in T v)
			if (op == "+" || op == "-" || op == "%") {
		mixin("value " ~ op ~ "= v.value;");
		return this;
	}
	DimensionedValue opOpAssign(string op)(in BaseType v) {
		mixin("value " ~ op ~ "= v;");
		return this;
	}

	private {
		template MulValueType(T) if (isDimensionedValue!(T)) {
			alias typeof(value * T.BaseType.init) MulValueType;
		}
		template MulValueType(T) if (!isDimensionedValue!(T)) {
			alias typeof(value * T.init) MulValueType;
		}

		template DivValueType(T) if (isDimensionedValue!(T)) {
			alias typeof(value / T.BaseType.init) DivValueType;
		}
		template DivValueType(T) if (!isDimensionedValue!(T)) {
			alias typeof(value / T.init) DivValueType;
		}
	}
}


private {
	template CheckDim(int Dim1, Dims...) {
		const bool CheckDim = (Dim1 != 0 || Dims.length != 0) && CheckDim!(Dims);
	}

	template CheckDim() {
		const bool CheckDim = true;
	}

	template DimOps() {
		template mul(Dims...) {
			alias Dims mul;
		}
		template div(Dims...) {
			alias DimOps!(Dims).inv div;
		}
		alias TypeTuple!() inv;
		alias TypeTuple!() id;
	}

	template DimOps(int Dim1, Dims1...) {
		template mul() {
			alias id mul;
		}
		template mul(int Dim2, Dims2...) {
			alias TypeTuple!(Dim1 + Dim2, DimOps!(Dims1).mul!(Dims2)) mul;
		}

		template div() {
			alias id div;
		}
		template div(int Dim2, Dims2...) {
			alias TypeTuple!(Dim1 - Dim2, DimOps!(Dims1).div!(Dims2)) div;
		}

		alias TypeTuple!(-Dim1, DimOps!(Dims1).inv) inv;
		// Bugzilla 2257 comment 7
		alias TypeTuple!(Dim1 + 0, DimOps!(Dims1).id) id;
	}

	template NormDim(Dims...) {
		static if (Dims.length == 0 || Dims[$-1] != 0) {
			alias Dims NormDim;
		} else {
			alias NormDim!(Dims[0..$-1]) NormDim;
		}
	}
}


unittest {
	import std.stdio;

	alias DimensionedValue!(int, 1) Length;
	alias DimensionedValue!(int, 0, 1) Time;

	Length x = Length(42);
	//x.value = 42;
	Length y = x + x;
	Length z = x + y;
	assert (x.value == 42);
	assert (y.value == 84);
	assert (z.value == 126);
	x -= z;
	assert (x.value == -84);
	x *= 2;
	y *= -1;
	z /= 5;
	assert (x.value == -168);
	assert (y.value == -84);
	assert (z.value == 25);

	x = Length(105);

	auto x2 = x * 2;
	auto x25 = x * 2.5;
	auto area = x * y;
	static assert (is(typeof(x2) == Length));
	static assert (is(typeof(x25) == DimensionedValue!(double, 1)));
	static assert (is(typeof(area) == DimensionedValue!(int, 2)));
	assert (x2 == 2 * x);
	assert (area.value == 105 * -84);

	auto freq = 42.5 / Time(2);
	auto speed = x * freq;
	//writefln("%s %s", freq.value, speed.value);
	auto distance = speed * Time(10);
	static assert (is(typeof(freq) == DimensionedValue!(double, 0, -1)));
	static assert (is(typeof(speed) == DimensionedValue!(double, 1, -1)));
	static assert (is(typeof(distance) == DimensionedValue!(double, 1)));
	assert (freq.value == 21.25);
	assert (speed.value == 2231.25);

	auto time = distance / speed;
	static assert (is(typeof(time) == DimensionedValue!(double, 0, 1)));
	static assert (is(typeof(time / time) == double));
	static assert (is(typeof(time * freq) == double));

	static assert (is(typeof(time / speed) == DimensionedValue!(double, -1, 2)));
	static assert (is(typeof(time / speed / area) == DimensionedValue!(double, -3, 2)));

	static assert (!is(typeof(time - speed)));
	static assert (!is(typeof(length + area)));

	//Length.MulType!(Frequency) test;
	//Length.DivType!(Time) test;
	//Frequency.DivRType!(Length) test;
	//pragma(msg, typeof(test));
	
	DimensionedValue!(real, 1) xr = x;
	assert (x.value == xr.value);
	assert (x == xr);
	assert (x <= xr);
	assert (x >= xr);
	assert (!(x < xr));
	assert (!(x > xr));
	assert (xr == x);
	assert (xr <= x);
	assert (xr >= x);
	assert (!(xr < x));
	assert (!(xr > x));
	xr += 20;
	assert (x < xr);
	assert (x <= xr);
	assert (x != xr);
	assert (!(x > xr));
	assert (!(x >= xr));
	assert (!(x == xr));
	assert (xr > x);
	assert (xr >= x);
	assert (xr != x);
	assert (!(xr < x));
	assert (!(xr <= x));
	assert (!(xr == x));

	puts("DimensionedValue passed!");
}
