/* ******************************************************************* *\
*                               hashset.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*           Hash table implementation of a set, which doesn't           *
*          assume that the key type has an ordering comparator          *
>***********************************************************************<
* Copyright (C) 2006-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
This module provides an implementation of a set data structure by means of a hash table.

Since: 0.03
*/
module smjg.libs.util.hashset;

private import
	smjg.libs.util.setbase,
	smjg.libs.util.types;

template HashSet(Element) if (!isImmutable!(Element) && isTailImmutable!(Element)) {
	alias HashSet!(immutable(Element)) HashSet;
}

///	The HashSet class template implements a set using a hash table.
class HashSet(Element) if (isImmutable!(Element)) : Set!(Element) {
	alias smjg.libs.util.setbase.Set!(Element) Set;

	public {
		/**	Constructs a new HashSet containing the specified elements.
		 *
		 *	Todo:
		 *		Templatise to support an array of implicitly convertible elements, once
		 *		$(BZ 435) is fixed.
		 */
		this(Element[] elements...) {
		//this(E)(E[] elements...) if (is(E : Element)) {
			foreach (elem; elements) add(elem);
		}

		override {
			///
			bool opIn_r(Element elem) const {
				const(Node*)* slot = typeid(Element).getHash(&elem) in hashTable;
				if (slot == null) return false;

				for (const(Node)* item = *slot; item != null; item = item.next) {
					if (item.key == elem) return true;
				}
				return false;
				/+const(Element[])* slot = typeid(Element).getHash(&elem) in hashTable;

				if (slot == null) return false;

				foreach (e; *slot) {
					if (e == elem) return true;
				}
				return false;+/
			}

			/**	Adds elem to the set.
			 *
			 *	Bugs:
			 *		As of DMD 2.058, it is possible to add a mutable object of a subclass of
			 *		Element's set ($(BZ 3731)).
			 */
			void add(Element elem) {
				hash_t hash = typeid(Element).getHash(&elem);
				Node** slot = hash in hashTable;

				if (slot == null) {
					hashTable[hash] = new Node(elem, null);
				} else {
					for (Node* item = *slot; item != null; item = item.next) {
						if (item.key == elem) return;
					}
					*slot = new Node(elem, *slot);
				}
			}

			///
			void remove(Element elem) {
				Node** slot = typeid(Element).getHash(&elem) in hashTable;
				if (slot == null) return;

				for (; *slot != null; slot = &(**slot).next) {
					if ((**slot).key == elem) {
						*slot = (**slot).next;
						return;
					}
				}
			}

			/**	Removes the element currently active in the innermost active foreach loop on this
			 *	set.
			 *
			 *	Precondition:
			 *		A foreach loop must be active on this set.
			 *	Warning:
			 *		The removal doesn't take effect until the iteration of the foreach loop exits.
			 *		A nested foreach loop on the same set must not be started after removeCurrent
			 *		before the end of the iteration, since this is likely to cause the wrong
			 *		element to be removed.  There is also potential undefined behaviour if the same
			 *		set is iterated through a mutable reference and, nested within this loop,
			 *		through a const or immutable reference.
			 *
			 *	Future_Directions:
			 *		This method may be superseded by something more robust.
			 */
			void removeCurrent()
			in {
				assert (inForEach != 0);
			}
			body {
				elementRemoved = true;
			}

			///
			@property Element storedKey(Element elem) const {
				const(Node*)* slot = typeid(Element).getHash(&elem) in hashTable;

				if (slot == null) return Element.init;

				for (const(Node)* item = *slot; item != null; item = item.next) {
					if (item.key == elem) return item.key;
				}
				return Element.init;
			}

			///
			@property size_t length() const {
				size_t result = 0;

				foreach (slot; hashTable) {
					for (const(Node)* item = slot; item != null; item = item.next) ++result;
				}
				return result;
			}

			///
			@property HashSet dup() const {
				HashSet result = new HashSet;

				foreach (e; this) result.add(e);
				result.hashTable.rehash;
				return result;
			}
			
			///
			@property immutable(HashSet) idup() const {
				return cast(immutable) dup();
			}

			///
			@property Element[] data() const {
				Element[] result;

				foreach (item; this) result ~= item;
				return result;
			}

			///
			int opApply(int delegate(ref Element) dg) {
				int result = 0;

				inForEach++;
				scope(exit) inForEach--;

				foreach (uint hash, ref slot; hashTable) {
					for (Node** item = &slot; *item != null;) {
						try {
							result = dg((**item).key);
						} finally {
							if (elementRemoved) {
								*item = (**item).next;
								elementRemoved = false;
							} else {
								item = &(**item).next;
							}
						}
						if (result) return result;
					}
				}
				return 0;
			}

			///
			int opApply(int delegate(ref Element) dg) const {
				int result = 0;
/*
				inForEach++;
				scope(exit) inForEach--;
*/
				foreach (uint hash, ref slot; hashTable) {
					for (const(Node*)* item = &slot; *item != null; item = &(**item).next) {
						result = dg((**item).key);
						if (result) return result;
					}
					/+for (uint index = 0; index < slot.length; index++) {
						//try {
							result = dg(cast(Element) slot[index]);
						/+} finally {
							if (elementRemoved) {
								slot[index] = slot[$-1];
								slot.length = slot.length - 1;
								index--;
							}
							elementRemoved = false;
						}+/
						if (result) return result;
					}+/
				}
				return 0;
			}

			/// TODO
			HashSet opAnd(const(Set) s) const {
				HashSet result = new HashSet;

				foreach (ref Element e; this) {
					if (e in s) result.add(e);
				}
				return result;
			}

			/*	Some of these are exactly the same as in the abstract Set
			 *	template, but are nonetheless duplicated for covariance.
			 */

			///
			HashSet opOr(const(Set) s) const {
				return dup.opOrAssign(s);
			}

			///
			HashSet opSub(const(Set) s) const {
				return dup.opSubAssign(s);
			}

			///
			HashSet opAnd(bool delegate(Element) dg) const {
				return dup.opAndAssign(dg);
			}

			///
			HashSet opSub(bool delegate(Element) dg) const {
				return dup.opSubAssign(dg);
			}

			///
			HashSet opXor(const(Set) s) const {
				return dup.opXorAssign(s);
			}

			///
			HashSet opAndAssign(const(Set) s) {
				foreach (ref Element e; this) {
					if (!(e in s)) remove(e);
				}
				return this;
			}

			///
			HashSet opOrAssign(const(Set) s) {
				foreach (ref Element e; s) add(e);
				return this;
			}

			///
			HashSet opSubAssign(const(Set) s) {
				super.opSubAssign(s);
				return this;
			}

			///
			HashSet opXorAssign(const(Set) s) {
				super.opXorAssign(s);
				return this;
			}

			///
			HashSet opAndAssign(bool delegate(Element) dg) {
				super.opAndAssign(dg);
				return this;
			}

			///
			HashSet opSubAssign(bool delegate(Element) dg) {
				super.opSubAssign(dg);
				return this;
			}
		}

		///	Rehashes the set in place.
		@property HashSet rehash() {
			HashSet newSet = dup;

			hashTable = newSet.hashTable;
			return this;
		}
	}

	private {
		struct Node {
			Element key;
			Node* next;
			this(Element k, Node* n) { key = k; next = n; }
		}

		//Element[][hash_t] hashTable;
		Node*[hash_t] hashTable;
		uint inForEach;
		bool elementRemoved;
	}
}

unittest {
	import std.stdio;

	alias HashSet!(int) IntSet;
	static assert (is(HashSet!(immutable(int)) == IntSet));

	IntSet s = new IntSet;

	assert (s.length == 0);

	s.add(42);
	s.add(66);
	s.add(66);
	assert (s.length == 2);
	assert (42 in s);
	assert (66 in s);
	s.remove(66);
	assert (s.length == 1);
	assert (!(66 in s));
	assert (!s[66]);
	assert (s[42]);

	int[100] data;
	foreach(uint i, ref int d; data) d = i;
	IntSet s1 = new IntSet(data.idup);
	IntSet s2 = s1 & delegate bool(immutable(int) i) { return i % 3 == 0; };
	s1 &= delegate bool(immutable(int) i) { return (i * i) % 5 == 4; };

	assert (s2.length == 34);
	assert (s1.length == 40);

	/*	Expected contents of s1 & s2:
	 *	3 12 18 27 33 42 48 57 63 72 78 87 93
	 */
	assert ((s1 & s2).length == 13);
	assert ((s1 | s2).length == 61);
	assert ((s1 - s2).length == 27);
	assert ((s1 ^ s2).length == 48);

	auto set2 = new HashSet!(immutable(HashSet!(immutable(int))));
	set2.add(s1.idup);
	set2.add(s2.idup);
	set2.add(s1.idup);

	puts("HashSet passed!");
}
